#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#ifdef HAVE_LIBENDEAVOUR2
# include <endeavour2.h>
#endif

#ifndef HAVE_LIBENDEAVOUR2
# include "../include/prochandle.h"
#endif

#include "cfg.h"

#include "guiutils.h"

#include "obj.h"
#include "obj_fio.h"
#include "avs_avengine_settings.h"
#include "avs_win.h"
#include "avs_win_animicon.h"
#include "avs_win_process.h"
#include "avs_win_cb.h"
#include "avs_win_dnd.h"
#include "avs_win_list.h"
#include "avs_win_db.h"
#include "avs_win_op_cb.h"
#include "avs_core.h"

#include "avs_cfg_list.h"
#include "config.h"

#include "images/icon_avscan_48x48.xpm"

#include "images/icon_avscan_20x20.xpm"
#include "images/icon_biohazard_20x20.xpm"
#include "images/icon_virus_db_20x20.xpm"
#include "images/icon_virus_db_32x32.xpm"
#include "images/icon_close_20x20.xpm"
#include "images/icon_exit_20x20.xpm"
#include "images/icon_add_20x20.xpm"
#include "images/icon_edit_20x20.xpm"
#include "images/icon_remove_20x20.xpm"
#include "images/icon_new_20x20.xpm"
#include "images/icon_cut_20x20.xpm"
#include "images/icon_copy_20x20.xpm"
#include "images/icon_paste_20x20.xpm"
#include "images/icon_cancel_20x20.xpm"
#include "images/icon_select_20x20.xpm"
#include "images/icon_properties_20x20.xpm"
#include "images/icon_start_20x20.xpm"
#include "images/icon_stop_20x20.xpm"
#include "images/icon_pause_20x20.xpm"
#include "images/icon_resume_20x20.xpm"
#include "images/icon_folder_opened_20x20.xpm"
#include "images/icon_tuning_20x20.xpm"
#include "images/icon_download_20x20.xpm"
#include "images/icon_trash_20x20.xpm"
#include "images/icon_move_file_20x20.xpm"
#include "images/icon_open_20x20.xpm"
#include "images/icon_save_as_20x20.xpm"
#include "images/icon_clear_20x20.xpm"
#include "images/icon_reload_20x20.xpm"
#include "images/icon_tool_bar_20x20.xpm"
#include "images/icon_status_bar_20x20.xpm"
#include "images/icon_help_20x20.xpm"
#include "images/icon_about_20x20.xpm"


/* Utilities */
gchar *avs_win_strdup_shorten(
	const gchar *s,
	const gint max_characters
);

/* Scan Location */
const gchar *avs_win_current_scan_location(AVSWin *win);
void avs_win_set_scan_location(
	AVSWin *win,
	const gchar *location
);
GList *avs_win_current_paths_list(AVSWin *win);

/* Scan Settings */
AVSScanSettings avs_win_current_scan_settings(AVSWin *win);
void avs_win_set_scan_settings(
	AVSWin *win,
	const AVSScanSettings options
);

/* Switch Page */
void avs_win_switch_page(
	AVSWin *win,
	const AVSWinPageIndex page_index
);

/* Scan List & Settings Update */
void avs_win_scan_list_open_file(AVSWin *win, const gchar *path);
void avs_win_scan_settings_update(
	AVSWin *win,
	GList *objs_list
);
void avs_win_set_results_location(AVSWin *win, GList *paths_list);
void avs_win_set_results_status(AVSWin *win, const gchar *s);

/* AVSWinOp */
AVSWinOp *avs_win_op_new(
	AVSWinOpID id,
	AVSWin *win
);
void avs_win_op_delete(AVSWinOp *opid);
GList *avs_win_ops_list_new(AVSWin *win);
void avs_win_ops_list_delete(GList *glist);
AVSWinOp *avs_win_ops_list_match_by_id(
	GList *ops_list,
	const AVSWinOpID id
);

/* AVScan Window */ 
static void avs_win_build_menu_bar(AVSWin *win);
static void avs_win_build_tool_bar(AVSWin *win);
static void avs_win_build_lists(AVSWin *win);
static void avs_win_build_status_bar(AVSWin *win);
static void avs_win_build_lists_menu(AVSWin *win);
AVSWin *avs_win_new(gpointer core_ptr);
void avs_win_sync_to_cfg_list(AVSWin *win);
void avs_win_update_title(AVSWin *win);
void avs_win_update_display(AVSWin *win);
void avs_win_set_busy(
	AVSWin *win,
	const gboolean busy
);
void avs_win_set_passive_busy(
	AVSWin *win,
	const gboolean busy
);
gboolean avs_win_tool_bar_is_shown(AVSWin *win);
void avs_win_show_tool_bar(
	AVSWin *win,
	const gboolean show
);
void avs_win_update_tool_bar(AVSWin *win);
gboolean avs_win_status_bar_is_shown(AVSWin *win);
void avs_win_show_status_bar(
	AVSWin *win,
	const gboolean show
);
void avs_win_set_status_progress(
	AVSWin *win,
	const gfloat v,
	const gboolean allow_gtk_iteration
);
void avs_win_set_status_message(
	AVSWin *win,
	const gchar *msg,
	const gboolean allow_gtk_iteration
);
void avs_win_set_status_time(
	AVSWin *win,
	const gulong t,
	const gboolean allow_gtk_iteration
);
gboolean avs_win_is_mapped(AVSWin *win);
void avs_win_map(AVSWin *win);
void avs_win_unmap(AVSWin *win);
void avs_win_delete(AVSWin *win);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : TRUE)


#define WIN_WIDTH			600
#define WIN_HEIGHT			400
#define WIN_STATUS_BAR_HEIGHT		26


/*
 *	Gets the current location as a single string containing a
 *	',' separated list of paths.
 */
const gchar *avs_win_current_scan_location(AVSWin *win)
{
	if(win == NULL)
		return(NULL);

	return(gtk_entry_get_text(GTK_ENTRY(win->scan_location_entry)));
}

/*
 *	Sets the current location.
 */
void avs_win_set_scan_location(
	AVSWin *win, const gchar *location
)
{
	GtkEntry *entry;

	if(win == NULL)
		return;

	entry = GTK_ENTRY(win->scan_location_entry);

	gtk_entry_set_text(
		entry,
		(location != NULL) ? location : ""
	);
	gtk_entry_set_position(entry, -1);
}

/*
 *	Gets the current paths list.
 *
 *	This is the equivilent to calling avs_win_current_scan_location()
 *	and using g_strsplit() to explode the ',' separated list of
 *	paths.
 *
 *	Returns a GList of gchar * paths, each item and the GList
 *	must be deleted by the calling function.
 */
GList *avs_win_current_paths_list(AVSWin *win)
{
	gchar **strv;
	const gchar *location;
	GList *glist;

	if(win == NULL)
		return(NULL);

	location = avs_win_current_scan_location(win);
	if(STRISEMPTY(location))
		return(NULL);

	strv = g_strsplit(location, ",", -1);
	glist = NULL;
	if(strv != NULL)
	{
		gint i;
		for(i = 0; strv[i] != NULL; i++)
			glist = g_list_append(glist, STRDUP(strv[i]));
		g_strfreev(strv);
	}

	return(glist);
}

/*
 *	Returns the current scan options.
 */
AVSScanSettings avs_win_current_scan_settings(AVSWin *win)
{
	AVSScanSettings options = 0;

	if(win == NULL)
		return(options);

	if(gtk_toggle_button_get_active(
		GTK_TOGGLE_BUTTON(win->scan_recursive_check)
	))
		options |= AVS_SCAN_RECURSIVE;
	if(gtk_toggle_button_get_active(
		GTK_TOGGLE_BUTTON(win->scan_executables_only_check)
	))
		options |= AVS_SCAN_EXECUTABLES_ONLY;
	if(gtk_toggle_button_get_active(
		GTK_TOGGLE_BUTTON(win->scan_ignore_links_check)
	))
		options |= AVS_SCAN_IGNORE_LINKS;

	return(options);
}

/*
 *	Sets the scan options.
 */
void avs_win_set_scan_settings(
	AVSWin *win,
	const AVSScanSettings options
)
{
	if(win == NULL)
		return;

	gtk_toggle_button_set_active(
		GTK_TOGGLE_BUTTON(win->scan_recursive_check),
		(options & AVS_SCAN_RECURSIVE) ? TRUE : FALSE
	);
	gtk_toggle_button_set_active(
		GTK_TOGGLE_BUTTON(win->scan_executables_only_check),
		(options & AVS_SCAN_EXECUTABLES_ONLY) ? TRUE : FALSE
	);
	gtk_toggle_button_set_active(
		GTK_TOGGLE_BUTTON(win->scan_ignore_links_check),
		(options & AVS_SCAN_IGNORE_LINKS) ? TRUE : FALSE
	);
}

/*
 *	Switch page.
 *
 *	The page_index specifies the page to switch to.
 *
 *	If you do not want the "switch_page" signal callback
 *	to do anything then you need to increment win->freeze_count
 *	> 0 before this call.
 */
void avs_win_switch_page(
	AVSWin *win,
	const AVSWinPageIndex page_index
)
{
	if(win == NULL)
		return;

	/* Do not freeze, so that avs_win_notebook_switch_page_cb() can
	 * perform any needed operations including the setting of
	 * win->current_page_index
	 */

	gtk_notebook_set_page(
		GTK_NOTEBOOK(win->notebook),
		(const gint)page_index
	);
}


/*
 *	Coppies the string any shortens it if it exceeds the maximum
 *	length.
 *
 *	The s specifies the string to copy.
 *
 *	The max_characters specifies the maximum length of characters.
 *
 *	Returns a copy of s that may have been shortened.
 */
gchar *avs_win_strdup_shorten(
	const gchar *s,
	const gint max_characters
)
{
	gint len;

	if(s == NULL)
		return(NULL);

	len = (gint)strlen((const char *)s);
	if((len > max_characters) && (max_characters > 3))
	{
		/* Need to shorten the string */
		const gint i = len - max_characters + 3;
		return(g_strconcat(
			"...",
			(const gchar *)(s + i),
			NULL
		));
	}
	else
	{
		return(g_strdup(s));
	}
}


/*
 *	Opens the scan list from file and updates the scan list.
 */
void avs_win_scan_list_open_file(AVSWin *win, const gchar *path)
{
	GList		*glist,
			*obj_list;
	GtkCList *clist;
#ifdef HAVE_LIBENDEAVOUR2
	EDVContext *edv_ctx;
#endif
	AVSCore *core;

	if(win == NULL)
		return;

	clist = GTK_CLIST(win->scan_clist);
	core = AVS_CORE(win->core);
#ifdef HAVE_LIBENDEAVOUR2
	edv_ctx = core->edv_ctx;
#endif

	/* Begin updating the scan list */

	gtk_clist_freeze(clist);

	/* Clear the scan list */
	gtk_clist_clear(clist);

	/* Open the scan list from file */
	obj_list = ObjListOpen(path, NULL, NULL);

	/* Transfer each object to the Win's scan list */
	for(glist = obj_list; glist != NULL; glist = g_list_next(glist))
		avs_win_scan_list_append(win, OBJ(glist->data));

	/* Delete scan list */
	g_list_free(obj_list);

	gtk_clist_thaw(clist);

	avs_win_update_display(win);
}

/*
 *	Updates the Win's scan settings with the values from the
 *	specified object.
 */
void avs_win_scan_settings_update(
	AVSWin *win,
	GList *objs_list
)
{
	gboolean b;
	gchar		*s,
			*scan_location;
	GList *glist;
	AVSScanSettings scan_options;
	obj_struct *obj;

	if(win == NULL)
		return;

	win->freeze_count++;

	/* Gather the scan settings from the objects list */
	scan_location = g_strdup("");
	scan_options = 0;
	for(glist = objs_list;
		glist != NULL;
		glist = g_list_next(glist)
	)
	{
		obj = OBJ(glist->data);
		if(obj == NULL)
			continue;

		s = g_strconcat(
			scan_location,
			obj->path,
			(g_list_next(glist) != NULL) ? "," : "",
			NULL
		);
		if(s != NULL)
		{
			g_free(scan_location);
			scan_location = s;
		}

		scan_options |= obj->options;
	}

	/* Set the scan settings */

	/* Scan location */
	gtk_entry_set_text(
		GTK_ENTRY(win->scan_location_entry),
		(scan_location != NULL) ? scan_location : ""
	);

	/* Recursive */
	b = (scan_options & AVS_SCAN_RECURSIVE) ? TRUE : FALSE;
	GUIMenuItemSetCheck(
		win->recursive_micheck,
		b,
		TRUE
	);
	gtk_toggle_button_set_active(
		GTK_TOGGLE_BUTTON(win->scan_recursive_check),
		b
	);

	/* Executables only */
	b = (scan_options & AVS_SCAN_EXECUTABLES_ONLY) ? TRUE : FALSE;
	GUIMenuItemSetCheck(
		win->executables_only_micheck,
		b,
		TRUE
	);
	gtk_toggle_button_set_active(
		GTK_TOGGLE_BUTTON(win->scan_executables_only_check),
		b
	);

	/* Ignore links */
	b = (scan_options & AVS_SCAN_IGNORE_LINKS) ? TRUE : FALSE;
	GUIMenuItemSetCheck(
		win->ignore_links_micheck,
		b,
		TRUE
	);
	gtk_toggle_button_set_active(
		GTK_TOGGLE_BUTTON(win->scan_ignore_links_check),
		b
	);

	avs_win_update_display(win);

	g_free(scan_location);

	win->freeze_count--;
}

/*
 *	Updates the Results Location.
 */
void avs_win_set_results_location(
	AVSWin *win,
	GList *paths_list
)
{
	gchar *s;

	if(win == NULL)
		return;

	if(paths_list != NULL)
	{
		const gint npaths = g_list_length(paths_list);
		if(npaths == 1)
		{
			const gchar *path = (const gchar *)paths_list->data;
			s = avs_win_strdup_shorten(path, 45);
		}
		else
		{
			GList *glist = g_list_last(paths_list);
			const gchar *last_path = (glist != NULL) ?
				(const gchar *)glist->data : "";
			gchar *st = g_strdup_printf(
				"..., %s (%i locations)",
				last_path,
				npaths
			);
			s = avs_win_strdup_shorten(st, 45);
			g_free(st);
		}
	}
	else
	{
		s = STRDUP("");
	}
	gtk_label_set_text(GTK_LABEL(win->results_location_label), s);
	g_free(s);
}

/*
 *	Updates the Results Status.
 */
void avs_win_set_results_status(AVSWin *win, const gchar *s)
{
	if(win == NULL)
		return;

	gtk_label_set_text(
		GTK_LABEL(win->results_status_label),
		(s != NULL) ? s : ""
	);
}


/*
 *	Creates a new Win OPID.
 */
AVSWinOp *avs_win_op_new(
	AVSWinOpID id, AVSWin *win
)
{
	AVSWinOp *opid = AVS_WIN_OP(
		g_malloc0(sizeof(AVSWinOp))
	);
	opid->id = id;
	opid->win = win;
	return(opid);
}

/*
 *	Deletes the Win OPID.
 */
void avs_win_op_delete(AVSWinOp *opid)
{
	if(opid == NULL)
		return;

	g_free(opid->button_name);
	g_free(opid->menu_name);
	g_free(opid->tooltip);
	g_free(opid);
}

/*
 *	Creates a new Win OPID list.
 */
GList *avs_win_ops_list_new(AVSWin *win)
{
	gboolean allow_multiple;
	ToolBarItemType item_type;
	const gchar	*button_name,
			*menu_name,
			*tooltip;
	guint		accel_key,
			accel_mods;
	guint8		**button_icon_data,
			**menu_icon_data;
	void (*func_cb)(GtkWidget *, gpointer);
	GList *glist = NULL;

#define DO_APPEND(_id_)	{			\
 AVSWinOp *opid = avs_win_op_new(		\
  (_id_), win					\
 );						\
 opid->allow_multiple = allow_multiple;		\
 opid->item_type = item_type;			\
 opid->button_name = STRDUP(button_name);	\
 opid->menu_name = STRDUP(menu_name);		\
 opid->tooltip = STRDUP(tooltip);		\
 opid->accel_key = accel_key;			\
 opid->accel_mods = accel_mods;			\
 opid->button_icon_data = button_icon_data;	\
 opid->menu_icon_data = menu_icon_data;		\
 opid->func_cb = func_cb;			\
 opid->enter_func_cb = avs_win_op_enter_event_cb;	\
 opid->leave_func_cb = avs_win_op_leave_event_cb;	\
 glist = g_list_append(glist, opid);		\
}

	allow_multiple = TRUE;
	item_type = TOOL_BAR_ITEM_SEPARATOR;
	button_name = NULL;
	menu_name = button_name;
	tooltip = NULL;
	accel_key = 0;
	accel_mods = 0;
	button_icon_data = NULL;
	menu_icon_data = button_icon_data;
	func_cb = NULL;
	DO_APPEND(AVS_WIN_OPID_SEPARATOR);


	allow_multiple = FALSE;
	item_type = TOOL_BAR_ITEM_BUTTON;
	button_name = "Close";
	menu_name = button_name;
	tooltip = "Close this window";
	accel_key = GDK_w;
	accel_mods = GDK_CONTROL_MASK;
	button_icon_data = (guint8 **)icon_close_20x20_xpm;
	menu_icon_data = button_icon_data;
	func_cb = WinCloseCB;
	DO_APPEND(AVS_WIN_OPID_CLOSE);

	allow_multiple = FALSE;
	item_type = TOOL_BAR_ITEM_BUTTON;
	button_name = "Exit";
	menu_name = "Exit";
	tooltip = "Close all of this application's windows and exit";
	accel_key = GDK_q;
	accel_mods = GDK_CONTROL_MASK; 
	button_icon_data = (guint8 **)icon_exit_20x20_xpm;
	menu_icon_data = button_icon_data;
	func_cb = WinExitCB;
	DO_APPEND(AVS_WIN_OPID_EXIT);


	allow_multiple = FALSE;
	item_type = TOOL_BAR_ITEM_BUTTON;
	button_name = "Add...";
	menu_name = "Add...";
	tooltip = "Add a new scan item";
	accel_key = 0;
	accel_mods = 0;
	button_icon_data = (guint8 **)icon_add_20x20_xpm;
	menu_icon_data = button_icon_data;
	func_cb = WinAddCB;
	DO_APPEND(AVS_WIN_OPID_ADD);

	allow_multiple = FALSE;
	item_type = TOOL_BAR_ITEM_BUTTON;
	button_name = "Edit...";
	menu_name = "Edit...";
	tooltip = "Edit the selected scan item";
	accel_key = 0;
	accel_mods = 0;
	button_icon_data = (guint8 **)icon_edit_20x20_xpm;
	menu_icon_data = button_icon_data;
	func_cb = WinEditCB;
	DO_APPEND(AVS_WIN_OPID_EDIT);

	allow_multiple = FALSE;
	item_type = TOOL_BAR_ITEM_BUTTON;
	button_name = "Remove";
	menu_name = "Remove";
	tooltip = "Remove the selected scan item(s)";
	accel_key = 0;
	accel_mods = 0;
	button_icon_data = (guint8 **)icon_remove_20x20_xpm;
	menu_icon_data = button_icon_data;
	func_cb = WinRemoveCB;
	DO_APPEND(AVS_WIN_OPID_REMOVE);

	allow_multiple = FALSE;
	item_type = TOOL_BAR_ITEM_BUTTON;
	button_name = "New...";
	menu_name = "New Scan Item...";
	tooltip = "Create a new scan item";
	accel_key = GDK_n;
	accel_mods = GDK_CONTROL_MASK;
	button_icon_data = (guint8 **)icon_new_20x20_xpm;
	menu_icon_data = button_icon_data;
	func_cb = WinAddCB;
	DO_APPEND(AVS_WIN_OPID_NEW);

	allow_multiple = FALSE;
	item_type = TOOL_BAR_ITEM_BUTTON;
	button_name = "Cut";
	menu_name = "Cut";
	tooltip = "Cut the selected scan item(s) to the clipboard";
	accel_key = GDK_x;
	accel_mods = GDK_CONTROL_MASK; 
	button_icon_data = (guint8 **)icon_cut_20x20_xpm;
	menu_icon_data = button_icon_data;
	func_cb = WinCutCB;
	DO_APPEND(AVS_WIN_OPID_CUT);

	allow_multiple = FALSE;
	item_type = TOOL_BAR_ITEM_BUTTON;
	button_name = "Copy";
	menu_name = "Copy";
	tooltip = "Copy the selected scan item(s) to the clipboard";
	accel_key = GDK_c;
	accel_mods = GDK_CONTROL_MASK; 
	button_icon_data = (guint8 **)icon_copy_20x20_xpm;
	menu_icon_data = button_icon_data;
	func_cb = WinCopyCB;
	DO_APPEND(AVS_WIN_OPID_COPY);

	allow_multiple = FALSE;
	item_type = TOOL_BAR_ITEM_BUTTON;
	button_name = "Paste";
	menu_name = "Paste";
	tooltip = "Paste scan item(s) from the clipboard";
	accel_key = GDK_v;
	accel_mods = GDK_CONTROL_MASK; 
	button_icon_data = (guint8 **)icon_paste_20x20_xpm;
	menu_icon_data = button_icon_data;
	func_cb = WinPasteCB;
	DO_APPEND(AVS_WIN_OPID_PASTE);

	allow_multiple = FALSE;
	item_type = TOOL_BAR_ITEM_BUTTON;
	button_name = "Delete";
	menu_name = "Delete";
	tooltip = "Delete the selected scan item(s)";
	accel_key = GDK_Delete;
	accel_mods = 0;
	button_icon_data = (guint8 **)icon_cancel_20x20_xpm;
	menu_icon_data = button_icon_data;
	func_cb = WinRemoveCB;
	DO_APPEND(AVS_WIN_OPID_DELETE);

	allow_multiple = FALSE;
	item_type = TOOL_BAR_ITEM_BUTTON;
	button_name = "SelAll";
	menu_name = "Select All";
	tooltip = "Select all scan items";
	accel_key = GDK_a;
	accel_mods = GDK_CONTROL_MASK;
	button_icon_data = (guint8 **)icon_select_20x20_xpm;
	menu_icon_data = NULL;
	func_cb = WinSelectAllCB;
	DO_APPEND(AVS_WIN_OPID_SELECT_ALL);

	allow_multiple = FALSE;
	item_type = TOOL_BAR_ITEM_BUTTON;
	button_name = "UnselAll";
	menu_name = "Unselect All";
	tooltip = "Unselect all scan items";
	accel_key = GDK_u;
	accel_mods = GDK_CONTROL_MASK;
	button_icon_data = (guint8 **)icon_select_20x20_xpm;
	menu_icon_data = NULL;
	func_cb = WinUnselectAllCB; 
	DO_APPEND(AVS_WIN_OPID_UNSELECT_ALL);

	allow_multiple = FALSE;
	item_type = TOOL_BAR_ITEM_BUTTON;
	button_name = "InvSel";
	menu_name = "Invert Selection";
	tooltip = "Invert selection of scan items";
	accel_key = GDK_i;
	accel_mods = GDK_CONTROL_MASK;
	button_icon_data = (guint8 **)icon_select_20x20_xpm;
	menu_icon_data = NULL;
	func_cb = WinInvertSelectionCB;
	DO_APPEND(AVS_WIN_OPID_INVERT_SELECTION);

	allow_multiple = FALSE;
	item_type = TOOL_BAR_ITEM_BUTTON;
	button_name = "Properties";
	menu_name = "Properties...";
	tooltip = "Edit the selected scan item's properties";
	accel_key = GDK_Return;
	accel_mods = GDK_CONTROL_MASK;
	button_icon_data = (guint8 **)icon_properties_20x20_xpm;
	menu_icon_data = button_icon_data;
	func_cb = WinEditCB;
	DO_APPEND(AVS_WIN_OPID_PROPERTIES);


	allow_multiple = FALSE;
	item_type = TOOL_BAR_ITEM_BUTTON;
	button_name = "Start";
	menu_name = button_name;
	tooltip = "Start scanning";
	accel_key = GDK_Return;
	accel_mods = 0;
	button_icon_data = (guint8 **)icon_start_20x20_xpm;
	menu_icon_data = button_icon_data;
	func_cb = WinStartCB;
	DO_APPEND(AVS_WIN_OPID_START);

	allow_multiple = FALSE;
	item_type = TOOL_BAR_ITEM_BUTTON;
	button_name = "Stop";
	menu_name = button_name;
	tooltip = "Stop the current operation";
	accel_key = GDK_Escape;
	accel_mods = 0;
	button_icon_data = (guint8 **)icon_stop_20x20_xpm;
	menu_icon_data = button_icon_data;
	func_cb = WinStopCB;
	DO_APPEND(AVS_WIN_OPID_STOP);

	allow_multiple = FALSE;
	item_type = TOOL_BAR_ITEM_BUTTON;
	button_name = "Pause";
	menu_name = button_name;
	tooltip = "Pause/resume the current operation";
	accel_key = GDK_p;
	accel_mods = GDK_CONTROL_MASK;
	button_icon_data = (guint8 **)icon_pause_20x20_xpm;
	menu_icon_data = button_icon_data;
	func_cb = WinPauseResumeCB;
	DO_APPEND(AVS_WIN_OPID_PAUSE_RESUME);

	allow_multiple = FALSE;
	item_type = TOOL_BAR_ITEM_BUTTON;
	button_name = "Location";
	menu_name = "Open Location...";
	tooltip = "Set the scan location";
	accel_key = GDK_o;
	accel_mods = GDK_CONTROL_MASK;
	button_icon_data = (guint8 **)icon_open_20x20_xpm;
	menu_icon_data = button_icon_data;
	func_cb = WinLocationCB;
	DO_APPEND(AVS_WIN_OPID_LOCATION);

	allow_multiple = FALSE;
	item_type = TOOL_BAR_ITEM_TOGGLE_BUTTON;
	button_name = "Recurse";
	menu_name = "Recursive";
	tooltip = "Toggle recursive scanning";
	accel_key = GDK_F6;
	accel_mods = 0;
	button_icon_data = NULL;
	menu_icon_data = button_icon_data;
	func_cb = WinRecursiveCB;
	DO_APPEND(AVS_WIN_OPID_RECURSIVE);

	allow_multiple = FALSE;
	item_type = TOOL_BAR_ITEM_TOGGLE_BUTTON;
	button_name = "ExecOnly";
	menu_name = "Executables Only";
	tooltip = "Toggle only scanning objects that are executable";
	accel_key = GDK_F7;
	accel_mods = 0;
	button_icon_data = NULL;
	menu_icon_data = button_icon_data;
	func_cb = WinExecutablesOnlyCB;
	DO_APPEND(AVS_WIN_OPID_EXECUTABLES_ONLY);

	allow_multiple = FALSE;
	item_type = TOOL_BAR_ITEM_TOGGLE_BUTTON;
	button_name = "IgLnks";
	menu_name = "Ignore Links";
	tooltip = "Toggle exclusion of scanning links";
	accel_key = GDK_F8;
	accel_mods = 0;
	button_icon_data = NULL;
	menu_icon_data = button_icon_data;
	func_cb = WinIgnoreLinksCB;
	DO_APPEND(AVS_WIN_OPID_IGNORE_LINKS);

	allow_multiple = FALSE;
	item_type = TOOL_BAR_ITEM_BUTTON;
	button_name = "ScanOpts";
	menu_name = "Scanner Settings...";
	tooltip = "Set the virus scanner's settings";
	accel_key = GDK_F9;
	accel_mods = 0;
	button_icon_data = (guint8 **)icon_tuning_20x20_xpm;
	menu_icon_data = button_icon_data;
	func_cb = WinScannerSettingsCB;
	DO_APPEND(AVS_WIN_OPID_SCANNER_SETTINGS);


	allow_multiple = FALSE;
	item_type = TOOL_BAR_ITEM_BUTTON;
	button_name = "Recycle";
	menu_name = "Recycle";
	tooltip = "Recycle the selected object(s)";
	accel_key = GDK_d;
	accel_mods = GDK_CONTROL_MASK;
	button_icon_data = (guint8 **)icon_trash_20x20_xpm;
	menu_icon_data = button_icon_data;
	func_cb = WinResultsRecycleCB;
	DO_APPEND(AVS_WIN_OPID_RESULTS_RECYCLE);

	allow_multiple = FALSE;
	item_type = TOOL_BAR_ITEM_BUTTON;
	button_name = "Move";
	menu_name = "Move...";
	tooltip = "Move the selected object(s)";
	accel_key = GDK_m;
	accel_mods = GDK_CONTROL_MASK;
	button_icon_data = (guint8 **)icon_move_file_20x20_xpm;
	menu_icon_data = button_icon_data;
	func_cb = WinResultsMoveCB;
	DO_APPEND(AVS_WIN_OPID_RESULTS_MOVE);

	allow_multiple = FALSE;
	item_type = TOOL_BAR_ITEM_BUTTON;
	button_name = "SelAll";
	menu_name = "Select All";
	tooltip = "Select all objects";
	accel_key = 0;
	accel_mods = 0;
	button_icon_data = (guint8 **)icon_select_20x20_xpm;
	menu_icon_data = NULL;
	func_cb = WinResultsSelectAllCB;
	DO_APPEND(AVS_WIN_OPID_RESULTS_SELECT_ALL);

	allow_multiple = FALSE;
	item_type = TOOL_BAR_ITEM_BUTTON;
	button_name = "UnselAll";
	menu_name = "Unselect All";
	tooltip = "Unselect all objects";
	accel_key = 0;
	accel_mods = 0;
	button_icon_data = (guint8 **)icon_select_20x20_xpm;
	menu_icon_data = NULL;
	func_cb = WinResultsUnselectAllCB;
	DO_APPEND(AVS_WIN_OPID_RESULTS_UNSELECT_ALL);

	allow_multiple = FALSE;
	item_type = TOOL_BAR_ITEM_BUTTON;
	button_name = "InvSel";
	menu_name = "Invert Selection";
	tooltip = "Invert selection of objects";
	accel_key = 0;
	accel_mods = 0;
	button_icon_data = (guint8 **)icon_select_20x20_xpm; 
	menu_icon_data = NULL;
	func_cb = WinResultsInvertSelectionCB;
	DO_APPEND(AVS_WIN_OPID_RESULTS_INVERT_SELECTION);

	allow_multiple = FALSE;
	item_type = TOOL_BAR_ITEM_BUTTON;
	button_name = "Report";
	menu_name = "Report...";
	tooltip = "View the scan report";
	accel_key = 0;
	accel_mods = 0;
	button_icon_data = (guint8 **)icon_edit_20x20_xpm;
	menu_icon_data = button_icon_data;
	func_cb = WinResultsReportCB;
	DO_APPEND(AVS_WIN_OPID_RESULTS_REPORT);

	allow_multiple = FALSE;
	item_type = TOOL_BAR_ITEM_BUTTON;
	button_name = "Open";
	menu_name = "Open...";
	tooltip = "Open a scan results log";
	accel_key = GDK_l;
	accel_mods = GDK_CONTROL_MASK;
	button_icon_data = (guint8 **)icon_open_20x20_xpm;
	menu_icon_data = button_icon_data;
	func_cb = WinResultsOpenCB;
	DO_APPEND(AVS_WIN_OPID_RESULTS_OPEN);

	allow_multiple = FALSE;
	item_type = TOOL_BAR_ITEM_BUTTON;
	button_name = "Save As";
	menu_name = "Save As...";
	tooltip = "Save the scan results as";
	accel_key = GDK_s;
	accel_mods = GDK_CONTROL_MASK;
	button_icon_data = (guint8 **)icon_save_as_20x20_xpm;
	menu_icon_data = button_icon_data;
	func_cb = WinResultsSaveAsCB;
	DO_APPEND(AVS_WIN_OPID_RESULTS_SAVE_AS);

	allow_multiple = FALSE;
	item_type = TOOL_BAR_ITEM_BUTTON;
	button_name = "Clear";
	menu_name = "Clear";
	tooltip = "Clear the scan results";
	accel_key = 0;
	accel_mods = 0;
	button_icon_data = (guint8 **)icon_clear_20x20_xpm;
	menu_icon_data = button_icon_data;
	func_cb = WinResultsClearCB;
	DO_APPEND(AVS_WIN_OPID_RESULTS_CLEAR);


	allow_multiple = FALSE;
	item_type = TOOL_BAR_ITEM_BUTTON;
	button_name = "Refresh";
	menu_name = "Refresh";
	tooltip = "Refresh the virus database";
	accel_key = GDK_F5;
	accel_mods = 0;
	button_icon_data = (guint8 **)icon_reload_20x20_xpm;
	menu_icon_data = button_icon_data;
	func_cb = WinDBRefreshCB;
	DO_APPEND(AVS_WIN_OPID_DB_REFRESH);

	allow_multiple = FALSE;
	item_type = TOOL_BAR_ITEM_BUTTON;
	button_name = "Location";
	menu_name = "Location...";
	tooltip = "Set the virus scanner's database location";
	accel_key = GDK_F10;
	accel_mods = 0;
	button_icon_data = (guint8 **)icon_virus_db_20x20_xpm;
	menu_icon_data = button_icon_data;
	func_cb = WinDBLocationCB;
	DO_APPEND(AVS_WIN_OPID_DB_LOCATION);

	allow_multiple = FALSE;
	item_type = TOOL_BAR_ITEM_BUTTON;
	button_name = "Update";
	menu_name = "Update From Internet";
	tooltip = "Get the latest virus database from the Internet";
	accel_key = GDK_F12;
	accel_mods = GDK_CONTROL_MASK;
	button_icon_data = (guint8 **)icon_download_20x20_xpm;
	menu_icon_data = button_icon_data;
	func_cb = avs_win_db_updateNetCB;
	DO_APPEND(AVS_WIN_OPID_DB_UPDATE_NET);


	allow_multiple = FALSE;
	item_type = TOOL_BAR_ITEM_TOGGLE_BUTTON;
	button_name = "Scan";
	menu_name = "Scan Page";
	tooltip = "Go to the scan page";
	accel_key = GDK_1;
	accel_mods = GDK_CONTROL_MASK;
	button_icon_data = NULL;
	menu_icon_data = NULL;
	func_cb = WinViewScanPageCB;
	DO_APPEND(AVS_WIN_OPID_VIEW_SCAN_PAGE);

	allow_multiple = FALSE;
	item_type = TOOL_BAR_ITEM_TOGGLE_BUTTON;
	button_name = "Results";
	menu_name = "Results Page"; 
	tooltip = "Go to the results page";
	accel_key = GDK_2;
	accel_mods = GDK_CONTROL_MASK;
	button_icon_data = NULL;
	menu_icon_data = NULL;
	func_cb = WinViewResultsPageCB;
	DO_APPEND(AVS_WIN_OPID_VIEW_RESULTS_PAGE);

	allow_multiple = FALSE;
	item_type = TOOL_BAR_ITEM_TOGGLE_BUTTON;
	button_name = "Database";
	menu_name = "Database Page";
	tooltip = "Go to the virus database page";
	accel_key = GDK_3;
	accel_mods = GDK_CONTROL_MASK;
	button_icon_data = NULL;
	menu_icon_data = NULL;
	func_cb = WinViewDBPageCB;
	DO_APPEND(AVS_WIN_OPID_VIEW_DB_PAGE);

	allow_multiple = FALSE;
	item_type = TOOL_BAR_ITEM_TOGGLE_BUTTON;
	button_name = "ToolBar";
	menu_name = "Tool Bar";
	tooltip = "Show/hide the tool bar";
	accel_key = 0;
	accel_mods = 0;
	button_icon_data = (guint8 **)icon_tool_bar_20x20_xpm;
	menu_icon_data = NULL;
	func_cb = WinViewToolBarCB;
	DO_APPEND(AVS_WIN_OPID_VIEW_TOOLBAR);

	allow_multiple = FALSE;
	item_type = TOOL_BAR_ITEM_TOGGLE_BUTTON;
	button_name = "StatusBar";
	menu_name = "Status Bar";
	tooltip = "Show/hide the status bar";
	accel_key = 0;
	accel_mods = 0;
	button_icon_data = (guint8 **)icon_status_bar_20x20_xpm;
	menu_icon_data = NULL;
	func_cb = WinViewStatusBarCB;
	DO_APPEND(AVS_WIN_OPID_VIEW_STATUSBAR);

	allow_multiple = FALSE;
	item_type = TOOL_BAR_ITEM_BUTTON;
	button_name = "Help";
	menu_name = "Contents";
	tooltip = "Index of all help pages";
	accel_key = GDK_F1;
	accel_mods = 0;
	button_icon_data = (guint8 **)icon_help_20x20_xpm;
	menu_icon_data = NULL;
	func_cb = WinHelpContentsCB;
	DO_APPEND(AVS_WIN_OPID_HELP_CONTENTS);

	allow_multiple = FALSE;
	item_type = TOOL_BAR_ITEM_BUTTON;
	button_name = "Scan?";
	menu_name = "How To Scan";
	tooltip = "Help on scanning objects for viruses";
	accel_key = 0;
	accel_mods = 0;
	button_icon_data = (guint8 **)icon_help_20x20_xpm;
	menu_icon_data = NULL;
	func_cb = WinHelpScanningCB;
	DO_APPEND(AVS_WIN_OPID_HELP_SCANNING);

	allow_multiple = FALSE;
	item_type = TOOL_BAR_ITEM_BUTTON;
	button_name = "Update?";
	menu_name = "How To Update";
	tooltip = "Help on updating the virus database";
	accel_key = 0;
	accel_mods = 0;
	button_icon_data = (guint8 **)icon_help_20x20_xpm;
	menu_icon_data = NULL;
	func_cb = WinHelpUpdatingCB;
	DO_APPEND(AVS_WIN_OPID_HELP_UPDATING);

	allow_multiple = FALSE;
	item_type = TOOL_BAR_ITEM_BUTTON;
	button_name = "About";
	menu_name = "About...";
	tooltip = "About this application";
	accel_key = 0;
	accel_mods = 0;
	button_icon_data = (guint8 **)icon_about_20x20_xpm;
	menu_icon_data = button_icon_data;
	func_cb = WinHelpAboutCB;
	DO_APPEND(AVS_WIN_OPID_ABOUT);


#undef DO_APPEND

	return(glist);
}

/*
 *	Deletes the Win OPID list.
 */
void avs_win_ops_list_delete(GList *glist)
{
	if(glist == NULL)
		return;

	g_list_foreach(glist, (GFunc)avs_win_op_delete, NULL);
	g_list_free(glist);
}

/*
 *	Returns the Win OPID that matches the specified id in the
 *	list.
 */
AVSWinOp *avs_win_ops_list_match_by_id(
	GList *ops_list,
	const AVSWinOpID id
)
{
	GList *glist;
	AVSWinOp *op;

	for(glist = ops_list;
	    glist != NULL;
	    glist = g_list_next(glist)
	)
	{
		op = AVS_WIN_OP(glist->data);
		if(op == NULL)
			continue;

		if(op->id == id)
			return(op);
	}

	return(NULL);
}



/*
 *	Builds the Win's Menu Bar.
 */
static void avs_win_build_menu_bar(AVSWin *win)
{
	guint8 **icon_data;
	GtkAccelGroup *accelgrp = win->accelgrp;
	GtkWidget	*w,
			*parent = win->menubar_handle,
			*menubar,
			*menu;

	/* Recreate the GtkMenuBar */
	GTK_WIDGET_DESTROY(win->menubar)
	win->menubar = menubar = w = GUIMenuBarCreate(NULL);
	gtk_widget_set_name(w, AVS_WIN_WIDGET_NAME_MENU_BAR);
	gtk_container_add(GTK_CONTAINER(parent), w);
	gtk_widget_show(w);

#define ADD_MENU_ITEM_LABEL(_id_)	{	\
 AVSWinOp *op = avs_win_ops_list_match_by_id(	\
  win->opids_list, (_id_)			\
 );						\
 if(op != NULL) {				\
  w = GUIMenuItemCreate(			\
   menu, GUI_MENU_ITEM_TYPE_LABEL, accelgrp,	\
   (icon_data != NULL) ?			\
    icon_data : op->menu_icon_data,		\
   op->menu_name,				\
   op->accel_key, op->accel_mods,		\
   op->func_cb, op->win				\
  );						\
  GUISetMenuItemCrossingCB(			\
   w,						\
   (gpointer)op->enter_func_cb, op,		\
   (gpointer)op->leave_func_cb, op		\
  );						\
 }						\
 else						\
  w = NULL;					\
}

#define ADD_MENU_ITEM_CHECK(_id_)	{	\
 AVSWinOp *op = avs_win_ops_list_match_by_id(	\
  win->opids_list, (_id_)			\
 );						\
 if(op != NULL) {				\
  w = GUIMenuItemCreate(			\
   menu, GUI_MENU_ITEM_TYPE_CHECK, accelgrp,	\
   (icon_data != NULL) ?			\
    icon_data : op->menu_icon_data,		\
   op->menu_name,				\
   op->accel_key, op->accel_mods,		\
   op->func_cb, op->win				\
  );						\
  GUISetMenuItemCrossingCB(			\
   w,						\
   (gpointer)op->enter_func_cb, op,		\
   (gpointer)op->leave_func_cb, op		\
  );						\
 }						\
 else                                           \
  w = NULL;					\
}

#define ADD_MENU_ITEM_SEPARATOR		{	\
 w = GUIMenuItemCreate(				\
  menu,						\
  GUI_MENU_ITEM_TYPE_SEPARATOR,			\
  NULL,						\
  NULL,						\
  NULL,						\
  0, 0,						\
  NULL, NULL                                    \
 );						\
}

	/* Create the Scan menu */
	menu = GUIMenuCreateTearOff();
	if(menu != NULL)
	{
		icon_data = NULL;
		ADD_MENU_ITEM_LABEL(AVS_WIN_OPID_NEW);
		win->new_mi = w;

		ADD_MENU_ITEM_SEPARATOR

		icon_data = NULL;
		ADD_MENU_ITEM_LABEL(AVS_WIN_OPID_LOCATION);
		win->location_mi = w;

		icon_data = NULL;
		ADD_MENU_ITEM_CHECK(AVS_WIN_OPID_RECURSIVE);
		win->recursive_micheck = w;

		icon_data = NULL;
		ADD_MENU_ITEM_CHECK(AVS_WIN_OPID_EXECUTABLES_ONLY);
		win->executables_only_micheck = w;

		icon_data = NULL;
		ADD_MENU_ITEM_CHECK(AVS_WIN_OPID_IGNORE_LINKS);
		win->ignore_links_micheck = w;

		ADD_MENU_ITEM_SEPARATOR

		icon_data = NULL;
		ADD_MENU_ITEM_LABEL(AVS_WIN_OPID_SCANNER_SETTINGS);
		win->scanner_options_mi = w;

		ADD_MENU_ITEM_SEPARATOR

		icon_data = NULL;
		ADD_MENU_ITEM_LABEL(AVS_WIN_OPID_START);
		win->start_mi = w;

		icon_data = NULL;
		ADD_MENU_ITEM_LABEL(AVS_WIN_OPID_STOP);
		win->stop_mi = w;

		icon_data = NULL;
		ADD_MENU_ITEM_LABEL(AVS_WIN_OPID_PAUSE_RESUME);
		win->pause_resume_mi = w;

		ADD_MENU_ITEM_SEPARATOR

		icon_data = NULL;
		ADD_MENU_ITEM_LABEL(AVS_WIN_OPID_CLOSE);
		win->close_mi = w;
	}
	win->scan_mlabel = GUIMenuAddToMenuBar(
		menubar,
		menu,
		"Scan",
		GTK_JUSTIFY_LEFT
	);

	/* Create the Edit menu */
	menu = GUIMenuCreateTearOff();
	if(menu != NULL)
	{
		icon_data = NULL;
		ADD_MENU_ITEM_LABEL(AVS_WIN_OPID_CUT);
		win->cut_mi = w;

		icon_data = NULL;
		ADD_MENU_ITEM_LABEL(AVS_WIN_OPID_COPY);
		win->copy_mi = w;

		icon_data = NULL;
		ADD_MENU_ITEM_LABEL(AVS_WIN_OPID_PASTE);
		win->paste_mi = w;

		icon_data = NULL;
		ADD_MENU_ITEM_LABEL(AVS_WIN_OPID_DELETE);
		win->delete_mi = w;

		ADD_MENU_ITEM_SEPARATOR

		icon_data = NULL;
		ADD_MENU_ITEM_LABEL(AVS_WIN_OPID_SELECT_ALL);
		icon_data = NULL;
		ADD_MENU_ITEM_LABEL(AVS_WIN_OPID_UNSELECT_ALL);
		icon_data = NULL;
		ADD_MENU_ITEM_LABEL(AVS_WIN_OPID_INVERT_SELECTION);

		ADD_MENU_ITEM_SEPARATOR

		icon_data = NULL;
		ADD_MENU_ITEM_LABEL(AVS_WIN_OPID_PROPERTIES);
		win->properties_mi = w;
	}
	win->edit_mlabel = GUIMenuAddToMenuBar(
		menubar, menu,
#if defined(PROG_LANGUAGE_SPANISH)
"Redacte"
#elif defined(PROG_LANGUAGE_FRENCH)
"Editer"
#elif defined(PROG_LANGUAGE_GERMAN)
"Redigieren"
#elif defined(PROG_LANGUAGE_ITALIAN)
"Redigere"
#elif defined(PROG_LANGUAGE_DUTCH)
"Bewerking"
#elif defined(PROG_LANGUAGE_PORTUGUESE)
"Edite"
#elif defined(PROG_LANGUAGE_NORWEGIAN)
"Rediger"
#else
"Edit"
#endif
		, GTK_JUSTIFY_LEFT
	);


	/* Create the View menu */
	menu = GUIMenuCreateTearOff();
	if(menu != NULL)
	{
		icon_data = NULL;
		ADD_MENU_ITEM_CHECK(AVS_WIN_OPID_VIEW_SCAN_PAGE);
		win->view_scan_page_micheck = w;

		icon_data = NULL;
		ADD_MENU_ITEM_CHECK(AVS_WIN_OPID_VIEW_RESULTS_PAGE);
		win->view_results_page_micheck = w;

		icon_data = NULL;
		ADD_MENU_ITEM_CHECK(AVS_WIN_OPID_VIEW_DB_PAGE);
		win->view_db_page_micheck = w;

		ADD_MENU_ITEM_SEPARATOR

		icon_data = NULL;
		ADD_MENU_ITEM_CHECK(AVS_WIN_OPID_VIEW_TOOLBAR);
		win->view_tool_bar_micheck = w;

		icon_data = NULL;
		ADD_MENU_ITEM_CHECK(AVS_WIN_OPID_VIEW_STATUSBAR);
		win->view_status_bar_micheck = w;
	}
	win->view_mlabel = GUIMenuAddToMenuBar( 
		menubar, menu,
#if defined(PROG_LANGUAGE_SPANISH)
"Vista"
#elif defined(PROG_LANGUAGE_FRENCH)
"Vue"
#elif defined(PROG_LANGUAGE_GERMAN)
"Blick"
#elif defined(PROG_LANGUAGE_ITALIAN)
"Veduta"
#elif defined(PROG_LANGUAGE_DUTCH)
"Overzicht"
#elif defined(PROG_LANGUAGE_PORTUGUESE)
"Vista"
#elif defined(PROG_LANGUAGE_NORWEGIAN)
"Sikt"
#else
"View"
#endif
		, GTK_JUSTIFY_LEFT
	);

	/* Create the Results menu */
	menu = GUIMenuCreateTearOff();
	if(menu != NULL)
	{
		icon_data = NULL;
		ADD_MENU_ITEM_LABEL(AVS_WIN_OPID_RESULTS_RECYCLE);
		win->results_recycle_mi = w;

		icon_data = NULL;
		ADD_MENU_ITEM_LABEL(AVS_WIN_OPID_RESULTS_MOVE);
		win->results_move_mi = w;

		ADD_MENU_ITEM_SEPARATOR

		icon_data = NULL;
		ADD_MENU_ITEM_LABEL(AVS_WIN_OPID_RESULTS_SELECT_ALL);
		win->results_select_all_mi = w;

		icon_data = NULL;
		ADD_MENU_ITEM_LABEL(AVS_WIN_OPID_RESULTS_UNSELECT_ALL);
		win->results_unselect_all_mi = w;

		icon_data = NULL;
		ADD_MENU_ITEM_LABEL(AVS_WIN_OPID_RESULTS_INVERT_SELECTION);
		win->results_invert_selection_mi = w;

		ADD_MENU_ITEM_SEPARATOR

		icon_data = NULL;
		ADD_MENU_ITEM_LABEL(AVS_WIN_OPID_RESULTS_REPORT);
		win->results_report_mi = w;

		ADD_MENU_ITEM_SEPARATOR

		icon_data = NULL;
		ADD_MENU_ITEM_LABEL(AVS_WIN_OPID_RESULTS_OPEN);
		win->results_open_mi = w;

		icon_data = NULL;
		ADD_MENU_ITEM_LABEL(AVS_WIN_OPID_RESULTS_SAVE_AS);
		win->results_save_as_mi = w;

		ADD_MENU_ITEM_SEPARATOR

		icon_data = NULL;
		ADD_MENU_ITEM_LABEL(AVS_WIN_OPID_RESULTS_CLEAR);
		win->results_clear_mi = w;
	}
	win->results_mlabel = GUIMenuAddToMenuBar(
		menubar,
		menu,
		"Results",
		GTK_JUSTIFY_LEFT
	);


	/* Create the Virus Database menu */
	menu = GUIMenuCreateTearOff();
	if(menu != NULL)
	{
		icon_data = NULL;
		ADD_MENU_ITEM_LABEL(AVS_WIN_OPID_DB_REFRESH);
		win->db_refresh_mi = w;

		ADD_MENU_ITEM_SEPARATOR

		icon_data = NULL;
		ADD_MENU_ITEM_LABEL(AVS_WIN_OPID_DB_LOCATION);
		win->db_location_mi = w; 

		ADD_MENU_ITEM_SEPARATOR

		icon_data = NULL;
		ADD_MENU_ITEM_LABEL(AVS_WIN_OPID_DB_UPDATE_NET);
		win->db_update_net_mi = w;
	}
	win->db_mlabel = GUIMenuAddToMenuBar(
		menubar,
		menu,
		"Database",
		GTK_JUSTIFY_LEFT
	);


	/* Create the Help menu */
	menu = GUIMenuCreateTearOff();
	if(menu != NULL)
	{
		icon_data = NULL;
		ADD_MENU_ITEM_LABEL(AVS_WIN_OPID_HELP_CONTENTS);
		win->help_contents_mi = w;

		ADD_MENU_ITEM_SEPARATOR

		icon_data = NULL;
		ADD_MENU_ITEM_LABEL(AVS_WIN_OPID_HELP_SCANNING);
/*		win->help_scanning_mi = w; */

		icon_data = NULL;
		ADD_MENU_ITEM_LABEL(AVS_WIN_OPID_HELP_UPDATING);
/*		win->help_scanning_mi = w; */

		ADD_MENU_ITEM_SEPARATOR

		icon_data = NULL;
		ADD_MENU_ITEM_LABEL(AVS_WIN_OPID_ABOUT);
		win->about_mi = w;
	}
	win->help_mlabel = GUIMenuAddToMenuBar( 
		menubar, menu,
#if defined(PROG_LANGUAGE_SPANISH)
"Ayuda"
#elif defined(PROG_LANGUAGE_FRENCH)
"Aide"
#elif defined(PROG_LANGUAGE_GERMAN)
"Hilfe"
#elif defined(PROG_LANGUAGE_ITALIAN)
"L'Aiuto"
#elif defined(PROG_LANGUAGE_DUTCH)
"Hulp"
#elif defined(PROG_LANGUAGE_PORTUGUESE)
"Ajuda"
#elif defined(PROG_LANGUAGE_NORWEGIAN)
"Hjelp"
#else
"Help"
#endif
		, GTK_JUSTIFY_RIGHT
	);

#undef ADD_MENU_ITEM_LABEL
#undef ADD_MENU_ITEM_CHECK
#undef ADD_MENU_ITEM_SEPARATOR
}

/*
 *	Builds the Win's Tool Bar.
 */
static void avs_win_build_tool_bar(AVSWin *win)
{
	const gint border_minor = 2;
	GList		*glist,
					*tb_cfg,
					*items_list;
	GtkWidget	*w,
					*parent = win->tool_bar_handle;
	ToolBarItem *item;
	ToolBarButtonDecals tb_display = TOOL_BAR_BUTTON_DECALS_PICTURES;
	GtkReliefStyle tb_relief = GTK_RELIEF_NONE;
	AVSWinOp *opid;

	/* Create the Tool Bar configuration */
	glist = NULL;
	glist = g_list_append(
		glist, (gpointer)AVS_WIN_OPID_START
	);
	glist = g_list_append(
		glist, (gpointer)AVS_WIN_OPID_STOP
	);
	glist = g_list_append(
		glist, (gpointer)AVS_WIN_OPID_PAUSE_RESUME
	);
	glist = g_list_append(
		glist, (gpointer)AVS_WIN_OPID_SCANNER_SETTINGS
	);
	glist = g_list_append(
		glist, (gpointer)AVS_WIN_OPID_SEPARATOR
	);
	glist = g_list_append(
		glist, (gpointer)AVS_WIN_OPID_NEW
	);
	glist = g_list_append(
		glist, (gpointer)AVS_WIN_OPID_CUT
	);
	glist = g_list_append(
		glist, (gpointer)AVS_WIN_OPID_COPY
	);
	glist = g_list_append(
		glist, (gpointer)AVS_WIN_OPID_PASTE
	);
	glist = g_list_append(
		glist, (gpointer)AVS_WIN_OPID_DELETE
	);
	glist = g_list_append(
		glist, (gpointer)AVS_WIN_OPID_PROPERTIES
	);
	glist = g_list_append(
		glist, (gpointer)AVS_WIN_OPID_SEPARATOR
	);
	glist = g_list_append(
		glist, (gpointer)AVS_WIN_OPID_RESULTS_RECYCLE
	);
	glist = g_list_append(
		glist, (gpointer)AVS_WIN_OPID_RESULTS_MOVE
	);
	glist = g_list_append(
		glist, (gpointer)AVS_WIN_OPID_RESULTS_REPORT
	);
	glist = g_list_append(
		glist, (gpointer)AVS_WIN_OPID_RESULTS_OPEN
	);
	glist = g_list_append(
		glist, (gpointer)AVS_WIN_OPID_RESULTS_SAVE_AS
	);
	glist = g_list_append(
		glist, (gpointer)AVS_WIN_OPID_RESULTS_CLEAR
	);
	glist = g_list_append(
		glist, (gpointer)AVS_WIN_OPID_SEPARATOR
	);
	glist = g_list_append(
		glist, (gpointer)AVS_WIN_OPID_DB_LOCATION
	);
	glist = g_list_append(
		glist, (gpointer)AVS_WIN_OPID_DB_UPDATE_NET
	);

	tb_cfg = glist;

	/* Create Tool Bar Items list from the Tool Bar configuration */
	items_list = NULL;
	for(glist = tb_cfg;
		glist != NULL;
		glist = g_list_next(glist)
	)
	{
		/* Get Operation ID from the Tool Bar configuration which
		 * specifies the ID
		 */
		opid = avs_win_ops_list_match_by_id(
			win->opids_list,
			(AVSWinOpID)glist->data
		);
		if(opid == NULL)
			continue;

		item = ToolBarItemNew(
			opid->item_type,		/* Item type */
			opid->button_name,		/* Label text */
			opid->button_icon_data,		/* Icon data */
			opid->tooltip,			/* Tooltip text */
			opid->id,			/* ID */
			avs_win_tool_bar_item_cb, opid,
			avs_win_tool_bar_item_enter_cb, opid,
			avs_win_tool_bar_item_leave_cb, opid
		);
		items_list = g_list_append(
			items_list,
			item
		);
	}

	/* Recreate the Tool Bar */
	GTK_WIDGET_DESTROY(win->tool_bar);
	win->tool_bar = w = ToolBarNew(
		GTK_ORIENTATION_HORIZONTAL,
		tb_display,
		tb_relief,
		GTK_POLICY_AUTOMATIC,
		items_list
	);
	gtk_widget_set_name(w, AVS_WIN_WIDGET_NAME_TOOL_BAR);
	gtk_container_add(GTK_CONTAINER(parent), w);
	gtk_container_border_width(GTK_CONTAINER(w), border_minor);
	gtk_widget_show(w);

	/* Delete the Tool Bar items list */
	if(items_list != NULL)
	{
		g_list_foreach(
			items_list,
			(GFunc)ToolBarItemDelete,
			NULL
		);
		g_list_free(items_list);
	}

	/* Delete the Tool Bar configuration */
	g_list_free(tb_cfg);
}

/*
 *	Builds the Win's Lists.
 */
static void avs_win_build_lists(AVSWin *win)
{
	const gint	border_major = 5,
			border_minor = 2;

	const GtkTargetEntry entry_dnd_tar_types[] = {
{GUI_TARGET_NAME_TEXT_PLAIN,	0,	DND_INFO_TEXT_PLAIN},
{GUI_TARGET_NAME_TEXT_URI_LIST,	0,	DND_INFO_TEXT_URI_LIST},
{GUI_TARGET_NAME_STRING,	0,	DND_INFO_STRING},
{AVS_DND_TARGET_SCAN_ITEM,	0,	AVS_DND_INFO_SCAN_ITEM}
	};
	const GtkTargetEntry scan_list_dnd_tar_types[] = {
{GUI_TARGET_NAME_TEXT_PLAIN,	0,	DND_INFO_TEXT_PLAIN},
{GUI_TARGET_NAME_TEXT_URI_LIST,	0,	DND_INFO_TEXT_URI_LIST},
{GUI_TARGET_NAME_STRING,	0,	DND_INFO_STRING},
{AVS_DND_TARGET_SCAN_ITEM,	0,	AVS_DND_INFO_SCAN_ITEM}
	};
	const GtkTargetEntry scan_list_dnd_src_types[] = {
{AVS_DND_TARGET_SCAN_ITEM,	0,	AVS_DND_INFO_SCAN_ITEM}
	};
	const GtkTargetEntry results_list_dnd_src_types[] = {
{GUI_TARGET_NAME_TEXT_PLAIN,	0,	DND_INFO_TEXT_PLAIN},
{GUI_TARGET_NAME_TEXT_URI_LIST,	0,	DND_INFO_TEXT_URI_LIST},
{GUI_TARGET_NAME_STRING,	0,	DND_INFO_STRING}
	};
	gint		font_size,
			columns;
	gchar		*font_name_h1_bold,
			**heading;
	const gchar *font_encoding;
	gpointer entry_rtn, browse_rtn;
	GdkFont *font;
	GtkRcStyle *rcstyle;
	GtkWidget	*w,
			*parent, *parent2, *parent3, *parent4,
			*parent5, *parent6, *parent7,
			*tablabel,
			*toplevel = win->toplevel;
	GtkNotebook *notebook;
	GtkCList *clist;
	GtkStyle *style = gtk_widget_get_style(toplevel);

#define NEW_ICON_BTN(_id_,_parent_,_pack_end_) {\
 AVSWinOp *opid = avs_win_ops_list_match_by_id(	\
  win->opids_list, (_id_)			\
 );						\
 if(opid != NULL) {				\
  w = GUIButtonPixmap(opid->button_icon_data);	\
  gtk_widget_set_usize(				\
   w,						\
   26,						\
   26						\
  );						\
  if(_pack_end_)				\
   gtk_box_pack_end(				\
    GTK_BOX(_parent_), w, FALSE, FALSE, 0	\
   );						\
  else						\
   gtk_box_pack_start(				\
    GTK_BOX(_parent_), w, FALSE, FALSE, 0	\
   );						\
  gtk_signal_connect(				\
   GTK_OBJECT(w), "enter_notify_event",		\
   GTK_SIGNAL_FUNC(opid->enter_func_cb), opid	\
  );						\
  gtk_signal_connect(				\
   GTK_OBJECT(w), "leave_notify_event",		\
   GTK_SIGNAL_FUNC(opid->leave_func_cb), opid	\
  );						\
  gtk_signal_connect(				\
   GTK_OBJECT(w), "clicked",			\
   GTK_SIGNAL_FUNC(opid->func_cb), opid->win	\
  );						\
  GUISetWidgetTip(w, opid->tooltip);		\
  gtk_widget_show(w);				\
 }						\
 else						\
  w = NULL;					\
}
#define NEW_ICON_LABEL_BTN(_id_,_parent_,_defaultable_){\
 AVSWinOp *opid = avs_win_ops_list_match_by_id(	\
  win->opids_list, (_id_)			\
 );						\
 if(opid != NULL) {				\
  w = GUIButtonPixmapLabelH(			\
   opid->button_icon_data, opid->button_name,	\
   NULL						\
  );						\
  if(_defaultable_) {				\
   GTK_WIDGET_SET_FLAGS(w, GTK_CAN_DEFAULT);	\
   gtk_widget_set_usize(			\
    w,						\
    GUI_BUTTON_HLABEL_WIDTH_DEF,		\
    GUI_BUTTON_HLABEL_HEIGHT_DEF		\
   );						\
  } else {					\
   gtk_widget_set_usize(			\
    w,						\
    GUI_BUTTON_HLABEL_WIDTH,			\
    GUI_BUTTON_HLABEL_HEIGHT			\
   );						\
  }						\
  gtk_box_pack_start(				\
   GTK_BOX(_parent_), w, FALSE, FALSE, 0	\
  );						\
  gtk_signal_connect(				\
   GTK_OBJECT(w), "enter_notify_event",		\
   GTK_SIGNAL_FUNC(opid->enter_func_cb), opid	\
  );						\
  gtk_signal_connect(				\
   GTK_OBJECT(w), "leave_notify_event",		\
   GTK_SIGNAL_FUNC(opid->leave_func_cb), opid	\
  );						\
  gtk_signal_connect(				\
   GTK_OBJECT(w), "clicked",			\
   GTK_SIGNAL_FUNC(opid->func_cb), opid->win	\
  );						\
  GUIButtonLabelUnderline(w, opid->accel_key);	\
  GUISetWidgetTip(w, opid->tooltip);		\
  gtk_widget_show(w);				\
 }						\
 else						\
  w = NULL;					\
}
#define SET_CROSSING(_id_,_w_)	{		\
 AVSWinOp *opid = avs_win_ops_list_match_by_id(	\
  win->opids_list, (_id_)			\
 );						\
 gtk_signal_connect(				\
  GTK_OBJECT(_w_), "enter_notify_event",	\
  GTK_SIGNAL_FUNC(opid->enter_func_cb), opid	\
 );						\
 gtk_signal_connect(				\
  GTK_OBJECT(_w_), "leave_notify_event",	\
  GTK_SIGNAL_FUNC(opid->leave_func_cb), opid	\
 );						\
}

	/* Get the base font size */
	font = style->font;
	font_size = GDK_FONT_GET_FONT_NAME_SIZE(font);
	if(font_size < 3)
	    font_size = 3;

	/* Format the font names */
#if defined(PROG_LANGUAGE_POLISH)
	font_encoding = "iso8859-2";
#else
	font_encoding = "iso8859-1";
#endif

	font_name_h1_bold = g_strdup_printf(
"-adobe-helvetica-bold-r-normal-*-%i-*-*-*-*-*-%s",
	    font_size + 2,
	    font_encoding
	);

	parent = win->lists_vbox;

	/* Main GtkHBox */
	w = gtk_hbox_new(FALSE, border_major);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_box_pack_start(GTK_BOX(parent), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent = w;
#if 0
	/* Left column vbox */
	w = gtk_vbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent2 = w;
#endif
	/* Left Column GtkNotebook */
	win->notebook = w = gtk_notebook_new();
	notebook = GTK_NOTEBOOK(w);
	gtk_widget_set_name(w, AVS_WIN_WIDGET_NAME_NOTEBOOK);
	gtk_box_pack_start(GTK_BOX(parent), w, TRUE, TRUE, 0); 
	gtk_notebook_set_tab_pos(notebook, GTK_POS_TOP);
	gtk_notebook_set_scrollable(notebook, TRUE);
	gtk_notebook_set_show_tabs(notebook, TRUE); 
	gtk_notebook_set_show_border(notebook, TRUE);
	gtk_signal_connect(
		GTK_OBJECT(w), "switch_page",
		GTK_SIGNAL_FUNC(avs_win_notebook_switch_page_cb), win
	);
	gtk_widget_show(w);
	parent2 = w;


	/* Scan page tab */
	tablabel = w = gtk_hbox_new(FALSE, border_minor);
	w = gtk_pixmap_new_from_xpm_d(
		NULL,
		NULL,
		(guint8 **)icon_avscan_20x20_xpm
	);
	gtk_box_pack_start(GTK_BOX(tablabel), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	w = gtk_label_new("Scan");
	gtk_box_pack_start(GTK_BOX(tablabel), w, FALSE, FALSE, 0);
	gtk_widget_show(w);

	/* Scan Page Toplevel GtkVBox */
	w = gtk_vbox_new(FALSE, border_major);
	gtk_widget_set_name(w, AVS_WIN_WIDGET_NAME_SCAN_PAGE_TOPLEVEL);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_notebook_append_page(
		GTK_NOTEBOOK(parent2), w, tablabel
	);
	gtk_widget_show(w);
	parent3 = w;

	/* Scan list frame */
	w = gtk_frame_new("Scan List");
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_box_pack_start(GTK_BOX(parent3), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent4 = w;

	w = gtk_vbox_new(FALSE, border_major);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_container_add(GTK_CONTAINER(parent4), w);
	gtk_widget_show(w);
	parent4 = w;

	/* Scan list scrolled window */
	w = gtk_scrolled_window_new(NULL, NULL); 
	gtk_scrolled_window_set_policy(
		GTK_SCROLLED_WINDOW(w),
		GTK_POLICY_NEVER,		/* No horizontal scroll bars */
		GTK_POLICY_AUTOMATIC
	);
	gtk_box_pack_start(GTK_BOX(parent4), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent5 = w;
	/* Scan list */
	columns = 3;
	heading = (gchar **)g_malloc(columns * sizeof(gchar *));
	heading[0] = "Name";
	heading[1] = "Location";
	heading[2] = "Last Scanned";
	win->scan_clist = w = gtk_clist_new_with_titles(columns, heading);
	g_free(heading);
	clist = GTK_CLIST(w);
	gtk_widget_set_name(w, AVS_WIN_WIDGET_NAME_SCAN_LIST);
/*	gtk_widget_set_usize(w, -1, 115); */
	gtk_widget_add_events(
		w,
		GDK_KEY_PRESS_MASK | GDK_KEY_RELEASE_MASK |
		GDK_ENTER_NOTIFY_MASK | GDK_LEAVE_NOTIFY_MASK |
		GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK |
		GDK_POINTER_MOTION_MASK | GDK_POINTER_MOTION_HINT_MASK
	);
	gtk_signal_connect(
		GTK_OBJECT(w), "key_press_event",
		GTK_SIGNAL_FUNC(avs_win_list_event_cb), win
	);
	gtk_signal_connect(
		GTK_OBJECT(w), "key_release_event",
		GTK_SIGNAL_FUNC(avs_win_list_event_cb), win
	);
	gtk_signal_connect(
		GTK_OBJECT(w), "button_press_event",
		GTK_SIGNAL_FUNC(avs_win_list_event_cb), win
	);
	gtk_signal_connect(
		GTK_OBJECT(w), "button_release_event",
		GTK_SIGNAL_FUNC(avs_win_list_event_cb), win
	);
	gtk_signal_connect(
		GTK_OBJECT(w), "motion_notify_event",
		GTK_SIGNAL_FUNC(avs_win_list_event_cb), win
	);
	gtk_signal_connect(
		GTK_OBJECT(w), "resize_column",
		GTK_SIGNAL_FUNC(WinListResizeColumnCB), win
	);
	gtk_signal_connect(
		GTK_OBJECT(w), "click_column",
		GTK_SIGNAL_FUNC(avs_win_list_click_column_cb), win
	);
	gtk_signal_connect(
		GTK_OBJECT(w), "select_row",
		GTK_SIGNAL_FUNC(avs_win_list_select_row_cb), win
	);
	gtk_signal_connect(
		GTK_OBJECT(w), "unselect_row",
		GTK_SIGNAL_FUNC(avs_win_list_unselect_row_cb), win
	);
	gtk_container_add(GTK_CONTAINER(parent5), w);
	gtk_widget_realize(w);
	gtk_clist_column_titles_passive(clist);
	gtk_clist_set_selection_mode(clist, GTK_SELECTION_EXTENDED);
	gtk_clist_set_column_auto_resize(clist, 0, TRUE);
	gtk_clist_set_column_resizeable(clist, 0, TRUE);
/*	gtk_clist_set_column_width(clist, 0, 140); */
	gtk_clist_set_column_justification(
		clist, 0, GTK_JUSTIFY_LEFT
	);
	gtk_clist_set_column_auto_resize(clist, 1, TRUE);
	gtk_clist_set_column_resizeable(clist, 1, TRUE);
/*	gtk_clist_set_column_width(clist, 1, 170); */
	gtk_clist_set_column_justification(
		clist, 1, GTK_JUSTIFY_LEFT
	);
	gtk_clist_set_column_auto_resize(clist, 2, TRUE);
	gtk_clist_set_column_resizeable(clist, 2, TRUE);
/*	gtk_clist_set_column_width(clist, 2, 10); */
	gtk_clist_set_column_justification(
		clist, 2, GTK_JUSTIFY_LEFT
	);
	gtk_clist_set_row_height(clist, AVS_WIN_LIST_ROW_SPACING);
	gtk_clist_set_shadow_type(clist, GTK_SHADOW_IN);
	GUIDNDSetSrc(
		w,
		scan_list_dnd_src_types,
		sizeof(scan_list_dnd_src_types) / sizeof(GtkTargetEntry),
		GDK_ACTION_COPY | GDK_ACTION_MOVE,	/* Actions */
		GDK_BUTTON1_MASK,			/* Buttons */
		NULL,
		avs_win_dnd_list_drag_data_get_cb,
		avs_win_dnd_list_drag_data_delete_cb,
		NULL,
		win
	); 
	GUIDNDSetTar(
		w,
		scan_list_dnd_tar_types,
		sizeof(scan_list_dnd_tar_types) / sizeof(GtkTargetEntry),
		GDK_ACTION_COPY | GDK_ACTION_MOVE,	/* Actions */ 
		GDK_ACTION_MOVE,			/* Default action if same */
		GDK_ACTION_COPY,			/* Default action */
		avs_win_dnd_list_drag_data_received_cb,
		win,
		TRUE				/* Highlight */
	); 
	gtk_widget_show(w);

	/* Scan list buttons hbox */
	w = gtk_hbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent5 = w;

	NEW_ICON_LABEL_BTN(AVS_WIN_OPID_ADD, parent5, FALSE);
	win->scan_add_btn = w;
	NEW_ICON_LABEL_BTN(AVS_WIN_OPID_EDIT, parent5, FALSE);
	win->scan_edit_btn = w;
	NEW_ICON_LABEL_BTN(AVS_WIN_OPID_REMOVE, parent5, FALSE);
	win->scan_remove_btn = w;


	/* Scan frame */
	w = gtk_frame_new("Scan");
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent4 = w;

	w = gtk_vbox_new(FALSE, border_major);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_container_add(GTK_CONTAINER(parent4), w);
	gtk_widget_show(w);
	parent4 = w;

	/* Location */
	w = gtk_hbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent5 = w;

	w = GUIPromptBarWithBrowse(
		NULL, "Location:",
		NULL, &entry_rtn, &browse_rtn,
		win, WinLocationCB
	);
	gtk_box_pack_start(GTK_BOX(parent5), w, TRUE, TRUE, 0);
	gtk_widget_show(w);

	w = (GtkWidget *)entry_rtn;
	win->scan_location_entry = w;
	GUIDNDSetTar(
		w,
		entry_dnd_tar_types,
		sizeof(entry_dnd_tar_types) / sizeof(GtkTargetEntry),
		GDK_ACTION_COPY,			/* Actions */
		GDK_ACTION_COPY,			/* Default action if same */
		GDK_ACTION_COPY,			/* Default action */
		avs_win_dnd_entry_drag_data_received_cb,
		win,
		TRUE				/* Highlight */
	);
	gtk_signal_connect(
		GTK_OBJECT(w), "key_press_event",
		GTK_SIGNAL_FUNC(avs_win_entry_complete_path_key_event_cb), win
	);
	gtk_signal_connect(
		GTK_OBJECT(w), "key_release_event",
		GTK_SIGNAL_FUNC(avs_win_entry_complete_path_key_event_cb), win
	);
	gtk_signal_connect(
		GTK_OBJECT(w), "changed",
		GTK_SIGNAL_FUNC(avs_win_location_entry_changed_cb), win
	);
	gtk_signal_connect(
		GTK_OBJECT(w), "activate",
		GTK_SIGNAL_FUNC(WinStartCB), win
	);
	SET_CROSSING(AVS_WIN_OPID_LOCATION, w);
	GUIEditableEndowPopupMenu(w, 0);
	GUISetWidgetTip(
		w,
"Enter a comma-separated list of full paths to the file(s)\
 and/or directory(ies) to scan"
	);

	w = (GtkWidget *)browse_rtn;
	win->scan_location_browse_btn = w;

	/* Scan settings hbox */
	w = gtk_hbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent5 = w;

	win->scan_recursive_check = w = gtk_check_button_new_with_label("Recursive");
	gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
	SET_CROSSING(AVS_WIN_OPID_RECURSIVE, w);
	GUISetWidgetTip(w,
"Check this to scan recursively into each sub directory (if the\
 specified location is a directory)"
	);
	gtk_widget_show(w);

	win->scan_executables_only_check = w = gtk_check_button_new_with_label(
		"Executables Only"
	);
	gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
	SET_CROSSING(AVS_WIN_OPID_EXECUTABLES_ONLY, w);
	GUISetWidgetTip(w,
"Check this to only scan objects who's permissions are set executable"
	);
	gtk_widget_show(w);

	win->scan_ignore_links_check = w = gtk_check_button_new_with_label(
		"Ignore Links"
	);
	gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
	SET_CROSSING(AVS_WIN_OPID_EXECUTABLES_ONLY, w);
	GUISetWidgetTip(w,
"Check this to exclude links from the scan"
	);
	gtk_widget_show(w);


	NEW_ICON_BTN(AVS_WIN_OPID_SCANNER_SETTINGS, parent5, TRUE);
	win->scan_scanner_options_btn = w;
	NEW_ICON_BTN(AVS_WIN_OPID_DB_LOCATION, parent5, TRUE);
	win->scan_db_location_btn = w;


	/* Results Page Tab */
	tablabel = w = gtk_hbox_new(FALSE, border_minor);
	w = gtk_pixmap_new_from_xpm_d(
		NULL,
		NULL,
		(guint8 **)icon_biohazard_20x20_xpm
	);
	gtk_box_pack_start(GTK_BOX(tablabel), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	w = gtk_label_new("Results");
	gtk_box_pack_start(GTK_BOX(tablabel), w, FALSE, FALSE, 0);
	gtk_widget_show(w);

	/* Results Page Toplevel GtkVBox */
	w = gtk_vbox_new(FALSE, border_major);
	gtk_widget_set_name(w, AVS_WIN_WIDGET_NAME_RESULTS_PAGE_TOPLEVEL);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_notebook_append_page(
		GTK_NOTEBOOK(parent2), w, tablabel
	);
	gtk_widget_show(w);
	parent3 = w;

	/* Results GtkFrame */
	w = gtk_frame_new("Results");
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_box_pack_start(GTK_BOX(parent3), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent4 = w;

	/* Results GtkVBox */
	w = gtk_vbox_new(FALSE, border_major);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_container_add(GTK_CONTAINER(parent4), w);
	gtk_widget_show(w);
	parent4 = w;

	/* Results row 1 GtkHBox */
	w = gtk_hbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent5 = w;

	/* Results left column GtkVBox */
	w = gtk_vbox_new(FALSE, border_minor);
	gtk_box_pack_start(GTK_BOX(parent5), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent6 = w;

	/* Results Location GtkHBox */
	w = gtk_hbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent6), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent7 = w;
	/* Prefix icon & label */
	w = gtk_pixmap_new_from_xpm_d(
		NULL,
		NULL,
		(guint8 **)icon_folder_opened_20x20_xpm
	);
	gtk_box_pack_start(GTK_BOX(parent7), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	w = gtk_label_new("Location:");
	gtk_label_set_justify(GTK_LABEL(w), GTK_JUSTIFY_RIGHT);
	gtk_box_pack_start(GTK_BOX(parent7), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	/* Results Location GtkLabel */
	win->results_location_label = w = gtk_label_new("");
	gtk_label_set_justify(GTK_LABEL(w), GTK_JUSTIFY_LEFT);
	gtk_box_pack_start(GTK_BOX(parent7), w, FALSE, FALSE, 0);
	gtk_widget_show(w);

	/* Results Status GtkHBox */
	w = gtk_hbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent6), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent7 = w;
	w = gtk_label_new("Status:");
	gtk_label_set_justify(GTK_LABEL(w), GTK_JUSTIFY_RIGHT);
	gtk_box_pack_start(GTK_BOX(parent7), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	/* Results Status GtkLabel */
	win->results_status_label = w = gtk_label_new("");
	gtk_label_set_justify(GTK_LABEL(w), GTK_JUSTIFY_LEFT);
	gtk_box_pack_start(GTK_BOX(parent7), w, FALSE, FALSE, 0);
	gtk_widget_show(w);


	/* Results right column buttons GtkVBox */
	w = gtk_vbox_new(TRUE, 0);
	gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent6 = w;

	NEW_ICON_LABEL_BTN(AVS_WIN_OPID_RESULTS_REPORT, parent6, FALSE);
	win->results_report_btn = w;


	/* Results List GtkScrolledWindow */
	w = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(
		GTK_SCROLLED_WINDOW(w),
		GTK_POLICY_AUTOMATIC,
		GTK_POLICY_AUTOMATIC
	);
	gtk_box_pack_start(GTK_BOX(parent4), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent5 = w;
	/* Results list */
	columns = 2;
	heading = (gchar **)g_malloc(columns * sizeof(gchar *));
	heading[0] = "Object";
	heading[1] = "Virus/Problem";
	win->results_clist = w = gtk_clist_new_with_titles(
		columns,
		heading
	);
	g_free(heading);
	clist = GTK_CLIST(w);
	gtk_widget_set_name(w, AVS_WIN_WIDGET_NAME_RESULTS_LIST);
	gtk_widget_add_events(
		w,
		GDK_KEY_PRESS_MASK | GDK_KEY_RELEASE_MASK |
		GDK_ENTER_NOTIFY_MASK | GDK_LEAVE_NOTIFY_MASK |
		GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK |
		GDK_POINTER_MOTION_MASK | GDK_POINTER_MOTION_HINT_MASK
	);
	gtk_signal_connect(
		GTK_OBJECT(w), "key_press_event",
		GTK_SIGNAL_FUNC(avs_win_list_event_cb), win
	);
	gtk_signal_connect(
		GTK_OBJECT(w), "key_release_event",
		GTK_SIGNAL_FUNC(avs_win_list_event_cb), win              
	);
	gtk_signal_connect(
		GTK_OBJECT(w), "button_press_event",
		GTK_SIGNAL_FUNC(avs_win_list_event_cb), win
	);
	gtk_signal_connect(
		GTK_OBJECT(w), "button_release_event",
		GTK_SIGNAL_FUNC(avs_win_list_event_cb), win
	);
	gtk_signal_connect(
		GTK_OBJECT(w), "motion_notify_event",
		GTK_SIGNAL_FUNC(avs_win_list_event_cb), win
	);
	gtk_signal_connect(
		GTK_OBJECT(w), "resize_column",
		GTK_SIGNAL_FUNC(WinListResizeColumnCB), win
	);
	gtk_signal_connect(
		GTK_OBJECT(w), "click_column",
		GTK_SIGNAL_FUNC(avs_win_list_click_column_cb), win
	);
	gtk_signal_connect(
		GTK_OBJECT(w), "select_row",
		GTK_SIGNAL_FUNC(avs_win_list_select_row_cb), win
	);
	gtk_signal_connect(
		GTK_OBJECT(w), "unselect_row",
		GTK_SIGNAL_FUNC(avs_win_list_unselect_row_cb), win  
	);
	gtk_container_add(GTK_CONTAINER(parent5), w);
	gtk_widget_realize(w);
	gtk_clist_column_titles_active(clist);
	gtk_clist_set_selection_mode(clist, GTK_SELECTION_EXTENDED);
	gtk_clist_set_column_auto_resize(clist, 0, TRUE);
	gtk_clist_set_column_resizeable(clist, 0, TRUE);
/*	gtk_clist_set_column_width(clist, 0, 220); */
	gtk_clist_set_column_justification(
		clist,
		0,
		GTK_JUSTIFY_LEFT
	);
	gtk_clist_set_column_auto_resize(clist, 1, TRUE);
	gtk_clist_set_column_resizeable(clist, 1, TRUE);
	gtk_clist_set_column_justification(
		clist,
		1,
		GTK_JUSTIFY_LEFT
	);
	gtk_clist_set_row_height(clist, AVS_WIN_LIST_ROW_SPACING);
	gtk_clist_set_shadow_type(clist, GTK_SHADOW_IN);
	GUIDNDSetSrc(
		w,
		results_list_dnd_src_types,
		sizeof(results_list_dnd_src_types) / sizeof(GtkTargetEntry),
		GDK_ACTION_COPY | GDK_ACTION_MOVE |
			GDK_ACTION_LINK,	/* Actions */
		GDK_BUTTON1_MASK,		/* Buttons */
		NULL,
		avs_win_dnd_list_drag_data_get_cb,
		avs_win_dnd_list_drag_data_delete_cb,
		NULL,
		win
	);
	gtk_widget_show(w);

	/* Results list buttons GtkHBox */
	w = gtk_hbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent5 = w;

	NEW_ICON_LABEL_BTN(AVS_WIN_OPID_RESULTS_RECYCLE, parent5, FALSE);
	win->results_recycle_btn = w;
	NEW_ICON_LABEL_BTN(AVS_WIN_OPID_RESULTS_MOVE, parent5, FALSE);
	win->results_move_btn = w;


	NEW_ICON_BTN(AVS_WIN_OPID_RESULTS_SAVE_AS, parent5, TRUE);
	win->results_save_as_btn = w;
	NEW_ICON_BTN(AVS_WIN_OPID_RESULTS_OPEN, parent5, TRUE);
	win->results_open_btn = w;
	NEW_ICON_BTN(AVS_WIN_OPID_RESULTS_CLEAR, parent5, TRUE);
	win->results_clear_btn = w;


	/* Virus Database Page Tab */
	tablabel = w = gtk_hbox_new(FALSE, border_minor);
	w = gtk_pixmap_new_from_xpm_d(
		NULL,
		NULL,
		(guint8 **)icon_virus_db_20x20_xpm
	);
	gtk_box_pack_start(GTK_BOX(tablabel), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	w = gtk_label_new("Database");
	gtk_box_pack_start(GTK_BOX(tablabel), w, FALSE, FALSE, 0);
	gtk_widget_show(w);

	/* Virus Database Page Toplevel GtkVBox*/
	w = gtk_vbox_new(FALSE, border_major);
	gtk_widget_set_name(w, AVS_WIN_WIDGET_NAME_DATABASE_PAGE_TOPLEVEL);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_notebook_append_page(
		GTK_NOTEBOOK(parent2), w, tablabel
	);
	gtk_widget_show(w);
	parent3 = w;

	/* Virus Database GtkFrame */
	w = gtk_frame_new("Virus Database");
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_box_pack_start(GTK_BOX(parent3), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent4 = w;

	w = gtk_vbox_new(FALSE, border_major);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_container_add(GTK_CONTAINER(parent4), w);
	gtk_widget_show(w);
	parent4 = w;

#define NEW_LABEL(_parent_,_label_,_w_rtn_) {		\
 GtkWidget *_parent2;					\
							\
 w = gtk_hbox_new(FALSE, border_minor);			\
 gtk_box_pack_start(GTK_BOX(_parent_), w, FALSE, FALSE, 0); \
 gtk_widget_show(w);					\
 _parent2 = w;						\
							\
 w = gtk_label_new(_label_);				\
 gtk_label_set_justify(GTK_LABEL(w), GTK_JUSTIFY_RIGHT);\
 gtk_box_pack_start(GTK_BOX(_parent2), w, FALSE, FALSE, 0);\
 gtk_widget_show(w);					\
							\
 (_w_rtn_) = w = gtk_label_new("");			\
 gtk_label_set_justify(GTK_LABEL(w), GTK_JUSTIFY_LEFT);	\
 gtk_box_pack_start(GTK_BOX(_parent2), w, FALSE, FALSE, 0);\
 gtk_widget_show(w);					\
							\
}
#define NEW_ENTRY(_parent_,_expand_,_label_,_w_rtn_,_units_) {	\
 GtkWidget *_parent2;					\
							\
 w = gtk_hbox_new(FALSE, border_minor);			\
 gtk_box_pack_start(GTK_BOX(_parent_), w, (_expand_), (_expand_), 0); \
 gtk_widget_show(w);					\
 _parent2 = w;						\
							\
 w = gtk_label_new(_label_);				\
 gtk_label_set_justify(GTK_LABEL(w), GTK_JUSTIFY_RIGHT);\
 gtk_box_pack_start(GTK_BOX(_parent2), w, FALSE, FALSE, 0);\
 gtk_widget_show(w);					\
							\
 (_w_rtn_) = w = gtk_entry_new();			\
 gtk_entry_set_editable(GTK_ENTRY(w), FALSE);		\
 gtk_box_pack_start(GTK_BOX(_parent2), w, TRUE, TRUE, 0);\
 gtk_widget_show(w);					\
							\
 if((_units_) != NULL) {				\
  w = gtk_label_new(_units_);				\
  gtk_label_set_justify(GTK_LABEL(w), GTK_JUSTIFY_LEFT);\
  gtk_box_pack_start(GTK_BOX(_parent2), w, FALSE, FALSE, 0);\
  gtk_widget_show(w);					\
 }							\
}

	w = gtk_hbox_new(FALSE, 2 * border_major);
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent5 = w;

	w = gtk_vbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent5), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent6 = w;

	/* Virus Database Name GtkHBox */
	w = gtk_hbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent6), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent7 = w;

	w = gtk_pixmap_new_from_xpm_d(
		NULL,
		NULL,
		(guint8 **)icon_virus_db_32x32_xpm
	);
	gtk_box_pack_start(GTK_BOX(parent7), w, FALSE, FALSE, 0);
	gtk_widget_show(w);

	win->db_name_label = w = gtk_label_new("");
	gtk_label_set_justify(GTK_LABEL(w), GTK_JUSTIFY_LEFT);
	gtk_box_pack_start(GTK_BOX(parent7), w, FALSE, FALSE, 0);
	rcstyle = gtk_rc_style_new();
	rcstyle->font_name = STRDUP(font_name_h1_bold);
	gtk_widget_modify_style(w, rcstyle);
	GTK_RC_STYLE_UNREF(rcstyle);
	gtk_widget_show(w);

	/* Location */
	NEW_ENTRY(
		parent6,
		FALSE,
		"Location:",
		win->db_location_entry,
		NULL
	);

	/* Updated */
	NEW_LABEL(
		parent6,
		"Last Updated:",
		win->db_updated_label
	);

	w = gtk_hbox_new(FALSE, 2 * border_major);
	gtk_box_pack_start(GTK_BOX(parent6), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent7 = w;

	/* Version */
	NEW_LABEL(
		parent7,
		"Version:",
		win->db_version_label
	);

	/* Timestamp */
	NEW_LABEL(
		parent7,
		"Build Date:",
		win->db_build_time_label
	);

	w = gtk_hbox_new(FALSE, 2 * border_major);
	gtk_box_pack_start(GTK_BOX(parent6), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent7 = w;

	/* Patterns */
	NEW_LABEL(
		parent7,
		"Patterns:",
		win->db_patterns_label
	);

	/* Size */
	NEW_LABEL(
		parent7,
		"Size:",
		win->db_size_label
	);

	/* Signature */
	NEW_ENTRY(
		parent6,
		FALSE,
		"Signature:",
		win->db_signature_entry,
		NULL
	);

	/* MD5 */
	NEW_ENTRY(
		parent6,
		FALSE,
		"MD5:",
		win->db_md5_entry,
		NULL
	);

	/* Builder */
	NEW_ENTRY(
		parent6,
		FALSE,
		"Builder:",
		win->db_builder_entry,
		NULL
	);

	w = gtk_hbox_new(FALSE, 2 * border_major);
	gtk_box_pack_start(GTK_BOX(parent6), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent7 = w;

	/* Functionality Level */
	NEW_LABEL(
		parent7,
		"Functionality Level:",
		win->db_functionality_level_label
	);


	w = gtk_vbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent6 = w;

	NEW_ICON_LABEL_BTN(AVS_WIN_OPID_DB_LOCATION, parent6, FALSE);
	win->db_location_btn = w;

	NEW_ICON_LABEL_BTN(AVS_WIN_OPID_DB_UPDATE_NET, parent6, FALSE);
	win->db_update_net_btn = w;


#undef NEW_LABEL
#undef NEW_ENTRY



	/* Right column GtkVBox */
	w = gtk_vbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent2 = w;


	/* Animated Icon GtkHBox */
	w = gtk_hbox_new(TRUE, 0);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent3 = w;
	/* Animated Icon */
	win->anim_icon = w = WinAnimIconNew();
	gtk_widget_set_name(w, AVS_WIN_WIDGET_NAME_ANIMATION_ICON);
	gtk_widget_add_events(
		w,
		GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK |
		GDK_POINTER_MOTION_MASK |
		GDK_ENTER_NOTIFY_MASK | GDK_LEAVE_NOTIFY_MASK
	);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	GUISetWidgetTip(
		w,
"Drag an object here to scan it for viruses"
	);
	GUIDNDSetTar(
		w,
		scan_list_dnd_tar_types,
		sizeof(scan_list_dnd_tar_types) / sizeof(GtkTargetEntry),
		GDK_ACTION_COPY,			/* Actions */
		GDK_ACTION_COPY,			/* Default action if same */
		GDK_ACTION_COPY,			/* Default action */
		avs_win_dnd_anim_icon_drag_data_received_cb,
		win,
		TRUE				/* Highlight */
	);
	gtk_widget_show(w);


	/* Buttons vbox */
	w = gtk_vbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent3 = w;

	NEW_ICON_LABEL_BTN(AVS_WIN_OPID_START, parent3, TRUE);
	win->start_btn = w;
	NEW_ICON_LABEL_BTN(AVS_WIN_OPID_STOP, parent3, TRUE);
	win->stop_btn = w;
	NEW_ICON_LABEL_BTN(AVS_WIN_OPID_PAUSE_RESUME, parent3, TRUE);
	win->pause_resume_btn = w;

	/* Buttons vbox */
	w = gtk_vbox_new(FALSE, 0);
	gtk_box_pack_end(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent3 = w;

	NEW_ICON_LABEL_BTN(AVS_WIN_OPID_CLOSE, parent3, TRUE);
	win->close_btn = w;


	g_free(font_name_h1_bold);

#undef NEW_ICON_BTN
#undef NEW_ICON_LABEL_BUTTON
#undef SET_CROSSING
}

/*
 *	Builds the Win's Status Bar.
 */
static void avs_win_build_status_bar(AVSWin *win)
{
	const gint border_minor = 2;
	gint attach_x = 0;
	GtkAdjustment *adj;
	GtkWidget	*w,
					*parent = win->main_vbox,
					*parent2, *parent3;

	/* Begin recreating the status bar */
	GTK_WIDGET_DESTROY(win->status_bar)

	/* Main outer frame */
	win->status_bar = w = gtk_frame_new(NULL);
	gtk_widget_set_name(w, AVS_WIN_WIDGET_NAME_STATUS_BAR);
	gtk_widget_set_usize(w, -1, WIN_STATUS_BAR_HEIGHT);
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_OUT);
	gtk_box_pack_end(GTK_BOX(parent), w, FALSE, FALSE, 0);
	parent = w;

	/* Main table */
	w = gtk_table_new(1, 3, FALSE);
	gtk_container_add(GTK_CONTAINER(parent), w);
	gtk_table_set_row_spacings(GTK_TABLE(w), border_minor);
	gtk_table_set_col_spacings(GTK_TABLE(w), border_minor);
	gtk_widget_show(w);
	parent = w;

	/* Progress bar */
	adj = (GtkAdjustment *)gtk_adjustment_new(
		0.0f, 1.0f, 100.0f,
		0.0f, 0.0f, 0.0f
	);
	win->status_bar_progress = w = gtk_progress_bar_new_with_adjustment(adj);
	gtk_widget_set_usize(w, 100, 20);
	gtk_progress_bar_set_orientation(
		GTK_PROGRESS_BAR(w), GTK_PROGRESS_LEFT_TO_RIGHT
	);
	gtk_progress_bar_set_bar_style(
		GTK_PROGRESS_BAR(w), GTK_PROGRESS_CONTINUOUS
	);
	gtk_progress_set_activity_mode(
		GTK_PROGRESS(w), FALSE
	);
	gtk_table_attach(
		GTK_TABLE(parent), w,
		attach_x, attach_x + 1,
		0, 1,
		0,
		GTK_SHRINK,
		0, 0
	);
	attach_x++;
	gtk_widget_show(w);


	/* Label */
	w = gtk_frame_new(NULL);
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_IN);
	gtk_table_attach(
		GTK_TABLE(parent), w,
		attach_x, attach_x + 1,
		0, 1,
		GTK_SHRINK | GTK_EXPAND | GTK_FILL,
		GTK_FILL,
		0, 0
	);
	attach_x++;
	gtk_widget_show(w);
	parent2 = w;

	w = gtk_hbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(parent2), w);
	gtk_widget_show(w);
	parent3 = w;

	win->status_bar_label = w = gtk_label_new("");
	gtk_label_set_justify(GTK_LABEL(w), GTK_JUSTIFY_LEFT);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, border_minor);
	gtk_widget_show(w);


	/* Time Label */
	w = gtk_frame_new(NULL);
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_IN);
	gtk_widget_set_usize(w, 60, -1);
	gtk_table_attach(
		GTK_TABLE(parent), w,
		attach_x, attach_x + 1,
		0, 1,
		0,
		GTK_FILL,
		0, 0
	);
	attach_x++;
	gtk_widget_show(w);
	parent2 = w;

	w = gtk_hbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(parent2), w);
	gtk_widget_show(w);
	parent3 = w;

	win->status_bar_time_label = w = gtk_label_new("");
	gtk_label_set_justify(GTK_LABEL(w), GTK_JUSTIFY_RIGHT);
	gtk_box_pack_end(GTK_BOX(parent3), w, FALSE, FALSE, border_minor);
	gtk_widget_show(w);
}


/*
 *	Builds the Win's right-click menus.
 */
static void avs_win_build_lists_menu(AVSWin *win)
{
	guint8 **icon_data;
	GtkAccelGroup *accelgrp = win->accelgrp;
	GtkWidget	*w,
					*menu;

#define ADD_MENU_ITEM_LABEL(_id_)	{	\
 AVSWinOp *opid = avs_win_ops_list_match_by_id(	\
  win->opids_list, (_id_)			\
 );						\
 if(opid != NULL) {				\
  w = GUIMenuItemCreate(			\
   menu, GUI_MENU_ITEM_TYPE_LABEL, accelgrp,	\
   (icon_data != NULL) ?			\
    icon_data : opid->menu_icon_data,		\
   opid->menu_name,				\
   opid->accel_key, opid->accel_mods,		\
   opid->func_cb, opid->win			\
  );						\
  GUISetMenuItemCrossingCB(			\
   w,						\
   (gpointer)opid->enter_func_cb, opid,		\
   (gpointer)opid->leave_func_cb, opid		\
  );						\
 }						\
 else                                           \
  w = NULL;					\
}

#define ADD_MENU_ITEM_CHECK(_id_)	{	\
 AVSWinOp *opid = avs_win_ops_list_match_by_id(	\
  win->opids_list, (_id_)			\
 );						\
 if(opid != NULL) {				\
  w = GUIMenuItemCreate(			\
   menu, GUI_MENU_ITEM_TYPE_CHECK, accelgrp,	\
   (icon_data != NULL) ?			\
    icon_data : opid->menu_icon_data,		\
   opid->menu_name,				\
   opid->accel_key, opid->accel_mods,		\
   opid->func_cb, opid->win			\
  );						\
  GUISetMenuItemCrossingCB(			\
   w,						\
   (gpointer)opid->enter_func_cb, opid,		\
   (gpointer)opid->leave_func_cb, opid		\
  );						\
 }						\
 else                                           \
  w = NULL;					\
}

#define ADD_MENU_ITEM_SEPARATOR		{	\
 w = GUIMenuItemCreate(				\
  menu,						\
  GUI_MENU_ITEM_TYPE_SEPARATOR,			\
  NULL,						\
  NULL,						\
  NULL,						\
  0, 0,						\
  NULL, NULL					\
 );						\
}
	/* Scan List right-click menu */
	GTK_WIDGET_DESTROY(win->scan_list_menu);
	win->scan_list_menu = menu = GUIMenuCreate();

	icon_data = NULL;
	ADD_MENU_ITEM_LABEL(AVS_WIN_OPID_START);
	win->scan_list_menu_start_mi = w;

	icon_data = NULL;
	ADD_MENU_ITEM_LABEL(AVS_WIN_OPID_STOP);
	win->scan_list_menu_stop_mi = w;

	icon_data = NULL;
	ADD_MENU_ITEM_LABEL(AVS_WIN_OPID_PAUSE_RESUME);
	win->scan_list_menu_pause_resume_mi = w;

	ADD_MENU_ITEM_SEPARATOR

	icon_data = NULL;
	ADD_MENU_ITEM_LABEL(AVS_WIN_OPID_NEW);
	win->scan_list_menu_new_mi = w;

	ADD_MENU_ITEM_SEPARATOR

	icon_data = NULL;
	ADD_MENU_ITEM_LABEL(AVS_WIN_OPID_CUT);
	win->scan_list_menu_cut_mi = w;

	icon_data = NULL;
	ADD_MENU_ITEM_LABEL(AVS_WIN_OPID_COPY);
	win->scan_list_menu_copy_mi = w;

	icon_data = NULL;
	ADD_MENU_ITEM_LABEL(AVS_WIN_OPID_PASTE);
	win->scan_list_menu_paste_mi = w;

	icon_data = NULL;
	ADD_MENU_ITEM_LABEL(AVS_WIN_OPID_DELETE);
	win->scan_list_menu_delete_mi = w;

	ADD_MENU_ITEM_SEPARATOR

	icon_data = NULL;
	ADD_MENU_ITEM_LABEL(AVS_WIN_OPID_SELECT_ALL);
	icon_data = NULL;
	ADD_MENU_ITEM_LABEL(AVS_WIN_OPID_UNSELECT_ALL);
	icon_data = NULL;
	ADD_MENU_ITEM_LABEL(AVS_WIN_OPID_INVERT_SELECTION);

	ADD_MENU_ITEM_SEPARATOR

	icon_data = NULL;
	ADD_MENU_ITEM_LABEL(AVS_WIN_OPID_PROPERTIES);
	win->scan_list_menu_properties_mi = w;


	/* Results List right-click menu */
	GTK_WIDGET_DESTROY(win->results_list_menu);
	win->results_list_menu = menu = GUIMenuCreate();

	icon_data = NULL;
	ADD_MENU_ITEM_LABEL(AVS_WIN_OPID_RESULTS_RECYCLE);
	win->results_list_menu_recycle_mi = w;

	icon_data = NULL;
	ADD_MENU_ITEM_LABEL(AVS_WIN_OPID_RESULTS_MOVE);
	win->results_list_menu_move_mi = w;

	ADD_MENU_ITEM_SEPARATOR

	icon_data = NULL;
	ADD_MENU_ITEM_LABEL(AVS_WIN_OPID_RESULTS_SELECT_ALL);
	win->results_list_menu_select_all_mi = w;

	icon_data = NULL;
	ADD_MENU_ITEM_LABEL(AVS_WIN_OPID_RESULTS_UNSELECT_ALL);
	win->results_list_menu_unselect_all_mi = w;

	icon_data = NULL;
	ADD_MENU_ITEM_LABEL(AVS_WIN_OPID_RESULTS_INVERT_SELECTION);
	win->results_list_menu_invert_selection_mi = w;


#undef ADD_MENU_ITEM_LABEL
#undef ADD_MENU_ITEM_CHECK
#undef ADD_MENU_ITEM_SEPARATOR
}



/*
 *	Creates a new Win.
 */
AVSWin *avs_win_new(gpointer core_ptr)
{
	GdkColormap *colormap;
	GdkWindow *window;
	GtkStyle *style;
	GtkAccelGroup *accelgrp;
	GtkWidget *w, *parent, *parent2;
	GtkWidget *toplevel, *main_vbox;
	CfgList *cfg_list;
	AVSCore *core = AVS_CORE(core_ptr);
	AVSWin *win = AVS_WIN(
		g_malloc0(sizeof(AVSWin))
	);

	cfg_list = core->cfg_list;

	win->toplevel = toplevel = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	win->accelgrp = accelgrp = gtk_accel_group_new();
/*
	win->processing = FALSE;
	win->paused = FALSE;
	win->busy_count = 0;
	win->passive_busy_count = 0;
	win->freeze_count = 0;
	win->stop_count = 0;
 */
	win->current_page_index = AVS_WIN_PAGE_SCAN;
	win->core = core_ptr;
	win->opids_list = avs_win_ops_list_new(win);
	win->menubar_map_state = TRUE;
	win->tool_bar_map_state = CFGItemListGetValueI(
		cfg_list, CFG_PARM_WIN_SHOW_TOOL_BAR
	);
	win->status_bar_map_state = CFGItemListGetValueI(
		cfg_list, CFG_PARM_WIN_SHOW_STATUS_BAR
	);
/*
	win->status_bar_progress_last = 0.0f;
	win->status_time = 0l;
 */
	win->scan_context = AVS_WIN_SCAN_CONTEXT(g_malloc0(
		sizeof(AVSWinScanContext)
	));
/*
	win->db_update_net_toid = 0;
	win->db_update_net_pid = 0;
	win->cstdout = NULL;
	win->cstderr = NULL;
	win->cstdout_line_buf = NULL;
	win->cstderr_line_buf = NULL;
 */

	/* Toplevel GtkWindow */
	w = toplevel;
	gtk_widget_set_name(w, AVS_WIN_WIDGET_NAME_TOPLEVEL);
	gtk_window_set_policy(GTK_WINDOW(w), FALSE, FALSE, FALSE);
	gtk_window_set_wmclass(
		GTK_WINDOW(w), "scanner", PROG_NAME
	);
	gtk_widget_add_events(
		w,
		GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK |
		GDK_KEY_PRESS_MASK | GDK_KEY_RELEASE_MASK
	);
	gtk_signal_connect(
		GTK_OBJECT(w), "key_press_event",
		GTK_SIGNAL_FUNC(avs_win_event_cb), win
	);
	gtk_signal_connect(
		GTK_OBJECT(w), "key_release_event",
		GTK_SIGNAL_FUNC(avs_win_event_cb), win
	);
	if(cfg_list != NULL)
	{
		const gint	x = CFGItemListGetValueI(cfg_list, CFG_PARM_WIN_X),
					y = CFGItemListGetValueI(cfg_list, CFG_PARM_WIN_Y);
		gtk_widget_set_uposition(w, x, y);
		gtk_widget_set_usize(w, WIN_WIDTH, WIN_HEIGHT);
	}
	gtk_widget_realize(w);
	window = w->window;
	if(window != NULL)
	{
		gdk_window_set_decorations(
			window,
			GDK_DECOR_BORDER | GDK_DECOR_TITLE | GDK_DECOR_MENU |
			GDK_DECOR_MINIMIZE
		);
		gdk_window_set_functions(
			window,
			GDK_FUNC_MOVE | GDK_FUNC_MINIMIZE | GDK_FUNC_CLOSE
		);
		GUISetWMIcon(window, (guint8 **)icon_avscan_48x48_xpm);
		gtk_window_set_title(GTK_WINDOW(w), PROG_NAME_FULL);
	}
	gtk_signal_connect(
		GTK_OBJECT(w), "delete_event",
		GTK_SIGNAL_FUNC(avs_win_event_cb), win
	);
	gtk_window_add_accel_group(GTK_WINDOW(w), accelgrp);
	parent = w;


	/* Styles */
#define REALLOC_COLOR(_ct_,_cs_)                {               \
 GDK_COLORMAP_FREE_COLOR(colormap, (_ct_));                     \
 memcpy((_ct_), (_cs_), sizeof(GdkColor));                      \
 GDK_COLORMAP_ALLOC_COLOR(colormap, (_ct_));                    \
}
	colormap = gdk_window_get_colormap(window);
	win->infected_style = style = gtk_style_copy(
		gtk_widget_get_style(w)
	);
	if(style != NULL)
	{
		const GdkColor	c_bg = { 0l, 0xffff, 0x0000, 0x0000 },
							c_fg = { 0l, 0xffff, 0xffff, 0xffff },
							c_bg2 = { 0l, 0xffff, 0x8080, 0x8080 },
							c_fg2 = { 0l, 0xa0a0, 0xa0a0, 0xa0a0 };

		GtkStateType state = GTK_STATE_NORMAL;
		REALLOC_COLOR(&style->fg[state], &c_fg);
		REALLOC_COLOR(&style->text[state], &c_fg);
		REALLOC_COLOR(&style->bg[state], &c_bg);
		REALLOC_COLOR(&style->base[state], &c_bg);

		state = GTK_STATE_PRELIGHT;
		REALLOC_COLOR(&style->fg[state], &c_fg);
		REALLOC_COLOR(&style->text[state], &c_fg);
		REALLOC_COLOR(&style->bg[state], &c_bg);
		REALLOC_COLOR(&style->base[state], &c_bg);

		state = GTK_STATE_ACTIVE;
		REALLOC_COLOR(&style->fg[state], &c_fg);
		REALLOC_COLOR(&style->text[state], &c_fg);
		REALLOC_COLOR(&style->bg[state], &c_bg);
		REALLOC_COLOR(&style->base[state], &c_bg);

		state = GTK_STATE_INSENSITIVE;
		REALLOC_COLOR(&style->fg[state], &c_fg2);
		REALLOC_COLOR(&style->text[state], &c_fg2);
		REALLOC_COLOR(&style->bg[state], &c_bg2);
		REALLOC_COLOR(&style->base[state], &c_bg2);
	}
#undef REALLOC_COLOR


	/* Main VBox */
	win->main_vbox = main_vbox = w = gtk_vbox_new(FALSE, 0);
	gtk_container_border_width(GTK_CONTAINER(w), 0);
	gtk_container_add(GTK_CONTAINER(parent), w);
	gtk_widget_show(w);
	parent = w;


	/* Build Menu Bar */
	win->menubar_handle = w = gtk_handle_box_new();
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	win->menubar = NULL;
	avs_win_build_menu_bar(win);

	/* Build Tool Bar */
	win->tool_bar_handle = w = gtk_handle_box_new();
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	win->tool_bar = NULL;
	avs_win_build_tool_bar(win);


	/* Vbox for the lists */
	win->lists_vbox = w = gtk_vbox_new(FALSE, 0);
/*	gtk_container_border_width(GTK_CONTAINER(w), 0); */
	gtk_box_pack_start(GTK_BOX(parent), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent2 = w;


	/* Build lists */
	avs_win_build_lists(win);


	/* Build status bar */
	win->status_bar = NULL;
	avs_win_build_status_bar(win);


	/* Build right click menu */
	win->scan_list_menu = NULL;
	avs_win_build_lists_menu(win);


	/* Show/hide menu bar */
	w = win->menubar_handle;
	if(win->menubar_map_state)
		gtk_widget_show(w);
	else
		gtk_widget_hide(w);

	/* Show/hide tool bar */
	w = win->tool_bar_handle;
	if(win->tool_bar_map_state)
		gtk_widget_show(w);
	else
		gtk_widget_hide(w);

	/* Show/hide status bar */
	w = win->status_bar;
	if(win->status_bar_map_state)
		gtk_widget_show(w);
	else
		gtk_widget_hide(w);


	/* Set the initial notebook page */
	gtk_notebook_set_page(
		GTK_NOTEBOOK(win->notebook), (gint)win->current_page_index
	);


	avs_win_update_display(win);


	return(win);
}

/*
 *	Records the Win's widget positions to the configuration.
 */
void avs_win_sync_to_cfg_list(AVSWin *win)
{
	GtkWidget *w;
	AVSCore *core = AVS_CORE((win != NULL) ? win->core : NULL);
	CfgItem *cfg_list = (core != NULL) ? core->cfg_list : NULL;

	if(cfg_list == NULL)
		return;

	/* Toplevel Position */
	w = win->toplevel;
	if(w != NULL)
	{
		GdkWindow *window = w->window;
		gint x = 0, y = 0;

		if(window != NULL)
			gdk_window_get_root_origin(window, &x, &y);

		CFGItemListSetValueI(
			cfg_list, CFG_PARM_WIN_X,
			x, FALSE
		);
		CFGItemListSetValueI(
			cfg_list, CFG_PARM_WIN_Y,
			y, FALSE
		);
#if 0
		width = w->allocation.width;
		height = w->allocation.height;
		CFGItemListSetValueI(
			cfg_list, CFG_PARM_WIN_WIDTH,
			width, FALSE
		);
		CFGItemListSetValueI(
			cfg_list, CFG_PARM_WIN_HEIGHT,
			height, FALSE
		);
#endif
	}

	/* Show Tool Bar */
	CFGItemListSetValueI(
		cfg_list, CFG_PARM_WIN_SHOW_TOOL_BAR,
		win->tool_bar_map_state, FALSE
	);

	/* Show Status Bar */
	CFGItemListSetValueI(
		cfg_list, CFG_PARM_WIN_SHOW_STATUS_BAR,
		win->status_bar_map_state, FALSE
	);

	/* Scan List */
	w = win->scan_clist;
	if(w != NULL)
	{
		gint i, n;
		gchar *path = g_strconcat(
			core->prog_data_dir,
			G_DIR_SEPARATOR_S,
			AVS_NAME_SCAN_LIST,
			NULL
		);
		GList *obj_list = NULL;
		GtkCList *clist = GTK_CLIST(w);

		/* Get the scan list */
		n = clist->rows;
		for(i = 0; i < n; i++)
			obj_list = g_list_append(
				obj_list,
				gtk_clist_get_row_data(clist, i)
			);

		/* Save scan list */
		ObjListSave(path, obj_list, NULL, NULL);

		g_list_free(obj_list);
	}

}

/*
 *	Updates the title to reflect current values.
 */
void avs_win_update_title(AVSWin *win)
{
	gchar *title;
	AVSWinScanContext *scan_ctx;

	if(win == NULL)
		return;

	scan_ctx = win->scan_context;

	/* Scanning? */
	if(scan_ctx->toid != 0)
	{
		if(scan_ctx->total > 0l)
		{
			const gfloat progress = (gfloat)scan_ctx->nscanned /
				(gfloat)scan_ctx->total;
			if(scan_ctx->ninfected > 0l)
				title = g_strdup_printf(
					PROG_NAME_FULL ": Scanning %i%% - %ld Infected%s",
					(gint)(progress * 100.0f),
					scan_ctx->ninfected,
					(win->paused) ? " (Paused)" : ""
				);
			else
				title = g_strdup_printf(
					PROG_NAME_FULL ": Scanning %i%%%s",
					(gint)(progress * 100.0f),
					(win->paused) ? " (Paused)" : ""
				);
		}
		else
		{
			title = g_strconcat(
				PROG_NAME_FULL ": Scanning",
				(win->paused) ? " (Paused)" : "",
				NULL
			);
		}
	}
	/* Updating Virus Database? */
	else if(win->db_update_net_toid != 0)
	{
		if(win->status_bar_progress_last < 0.0f)
			title = g_strconcat(
				PROG_NAME_FULL ": Updating Virus Database",
				(win->paused) ? " (Paused)" : "",
				NULL
			);
		else
			title = g_strdup_printf(
				PROG_NAME_FULL ": Updating Virus Database (%i%%)%s",
				(gint)(win->status_bar_progress_last * 100.0f),
				(win->paused) ? " (Paused)" : ""
			);
	}
	else
	{
		switch(win->current_page_index)
		{
		  case AVS_WIN_PAGE_SCAN:
			title = STRDUP(
				PROG_NAME_FULL
			);
			break;
		  case AVS_WIN_PAGE_RESULTS:
			title = STRDUP(
				PROG_NAME_FULL ": Results"
			);
			break;
		  case AVS_WIN_PAGE_DB:
			title = STRDUP(
				PROG_NAME_FULL ": Virus Database"
			);
			break;
		  default:
			title = STRDUP(
				PROG_NAME_FULL         
			);
			break;
		}
	}
	gtk_window_set_title(GTK_WINDOW(win->toplevel), title);
	g_free(title);
}

/*
 *	Updates the widgets to reflect current values.
 */
void avs_win_update_display(AVSWin *win)
{
	gboolean	sensitive,
			toggled,
			scanning,
			db_updating;
	const gchar	*s,
			*location_s;
	guint8 **icon_data;
	gint		last_selected_scan_item,
			last_selected_results_item,
			nselected_scan_items;
	AVSWinPageIndex page_num;
	GList *glist;
	GtkCList *clist;
	GtkWidget *tool_bar;
	AVSWinOp *opid;

	if(win == NULL)
		return;

	win->freeze_count++;

	tool_bar = win->tool_bar;

	page_num = win->current_page_index;
	scanning = avs_win_scan_process_is_running(win);
	db_updating = avs_win_db_update_net_process_is_running(win);
	location_s = gtk_entry_get_text(GTK_ENTRY(win->scan_location_entry));

	clist = GTK_CLIST(win->scan_clist);
	glist = (clist != NULL) ? clist->selection_end : NULL;
	last_selected_scan_item = (glist != NULL) ?
		(gint)glist->data : -1;
	nselected_scan_items = g_list_length(clist->selection);

	clist = GTK_CLIST(win->results_clist);
	glist = (clist != NULL) ? clist->selection_end : NULL;
	last_selected_results_item = (glist != NULL) ?           
		(gint)glist->data : -1;


	/* Menu Bar Labels */
	sensitive = TRUE;
	GTK_WIDGET_SET_SENSITIVE(win->scan_mlabel, sensitive);
	sensitive = (page_num == AVS_WIN_PAGE_SCAN) ? TRUE : FALSE;
	GTK_WIDGET_SET_SENSITIVE(win->edit_mlabel, sensitive);
	sensitive = TRUE;
	GTK_WIDGET_SET_SENSITIVE(win->view_mlabel, sensitive);
	sensitive = (page_num == AVS_WIN_PAGE_RESULTS) ? TRUE : FALSE;
	GTK_WIDGET_SET_SENSITIVE(win->results_mlabel, sensitive);
	sensitive = TRUE;
	GTK_WIDGET_SET_SENSITIVE(win->db_mlabel, sensitive);
	sensitive = TRUE;
	GTK_WIDGET_SET_SENSITIVE(win->help_mlabel, sensitive);

	/* New and Add */
	sensitive = (page_num == AVS_WIN_PAGE_SCAN) ? TRUE : FALSE;
	GTK_WIDGET_SET_SENSITIVE(win->new_mi, sensitive);
	GTK_WIDGET_SET_SENSITIVE(win->scan_add_btn, sensitive);
	ToolBarSetItemSensitive(
		tool_bar, AVS_WIN_OPID_NEW, sensitive
	);
	ToolBarSetItemSensitive(
		tool_bar, AVS_WIN_OPID_ADD, sensitive
	);

	/* Close */
	sensitive = (scanning || db_updating) ? FALSE : TRUE;
	GTK_WIDGET_SET_SENSITIVE(win->close_mi, sensitive);
	GTK_WIDGET_SET_SENSITIVE(win->close_btn, sensitive);
	ToolBarSetItemSensitive(
		tool_bar, AVS_WIN_OPID_CLOSE, sensitive
	);


	/* Cut, Copy, Delete, Edit, Remove, and Properties */
	sensitive = ((page_num == AVS_WIN_PAGE_SCAN) &&
		(last_selected_scan_item > -1)) ? TRUE : FALSE;
	GTK_WIDGET_SET_SENSITIVE(win->cut_mi, sensitive);
	GTK_WIDGET_SET_SENSITIVE(win->copy_mi, sensitive);
	GTK_WIDGET_SET_SENSITIVE(win->delete_mi, sensitive);
	GTK_WIDGET_SET_SENSITIVE(win->scan_remove_btn, sensitive);
	GTK_WIDGET_SET_SENSITIVE(win->scan_list_menu_cut_mi, sensitive);
	GTK_WIDGET_SET_SENSITIVE(win->scan_list_menu_copy_mi, sensitive);
	GTK_WIDGET_SET_SENSITIVE(win->scan_list_menu_delete_mi, sensitive);
	ToolBarSetItemSensitive(
		tool_bar, AVS_WIN_OPID_REMOVE, sensitive
	);
	ToolBarSetItemSensitive(
		tool_bar, AVS_WIN_OPID_CUT, sensitive
	);
	ToolBarSetItemSensitive(
		tool_bar, AVS_WIN_OPID_COPY, sensitive
	);
	ToolBarSetItemSensitive(
		tool_bar, AVS_WIN_OPID_DELETE, sensitive
	);

	/* Edit & Properties */
	sensitive = ((page_num == AVS_WIN_PAGE_SCAN) &&
		(last_selected_scan_item > -1) &&
		(nselected_scan_items == 1)
	) ? TRUE : FALSE;
	GTK_WIDGET_SET_SENSITIVE(win->properties_mi, sensitive);
	GTK_WIDGET_SET_SENSITIVE(win->scan_edit_btn, sensitive);
	GTK_WIDGET_SET_SENSITIVE(win->scan_list_menu_properties_mi, sensitive);
	ToolBarSetItemSensitive(
		tool_bar, AVS_WIN_OPID_EDIT, sensitive
	);
	ToolBarSetItemSensitive(
		tool_bar, AVS_WIN_OPID_PROPERTIES, sensitive
	);


	/* Paste */
	sensitive = (page_num == AVS_WIN_PAGE_SCAN) ? TRUE : FALSE;
	GTK_WIDGET_SET_SENSITIVE(win->paste_mi, sensitive);
	GTK_WIDGET_SET_SENSITIVE(win->scan_list_menu_paste_mi, sensitive);
	ToolBarSetItemSensitive(
		tool_bar, AVS_WIN_OPID_PASTE, sensitive
	);

	/* Start */
	sensitive = (STRISEMPTY(location_s) ||
		scanning || db_updating) ? FALSE : TRUE;
	GTK_WIDGET_SET_SENSITIVE(win->start_mi, sensitive);
	GTK_WIDGET_SET_SENSITIVE(win->start_btn, sensitive);
	GTK_WIDGET_SET_SENSITIVE(win->scan_list_menu_start_mi, sensitive);
	ToolBarSetItemSensitive(
		tool_bar, AVS_WIN_OPID_START, sensitive
	);

	/* Stop */
	sensitive = (scanning || db_updating) ? TRUE : FALSE;
	GTK_WIDGET_SET_SENSITIVE(win->stop_mi, sensitive);
	GTK_WIDGET_SET_SENSITIVE(win->stop_btn, sensitive);
	GTK_WIDGET_SET_SENSITIVE(win->scan_list_menu_stop_mi, sensitive);
	ToolBarSetItemSensitive(
		tool_bar, AVS_WIN_OPID_STOP, sensitive
	);

	/* Pause/Resume */
	sensitive = (scanning || db_updating) ? TRUE : FALSE;
	GTK_WIDGET_SET_SENSITIVE(win->pause_resume_mi, sensitive);
	GTK_WIDGET_SET_SENSITIVE(win->pause_resume_btn, sensitive);
	GTK_WIDGET_SET_SENSITIVE(win->scan_list_menu_pause_resume_mi, sensitive);
	ToolBarSetItemSensitive(
		tool_bar, AVS_WIN_OPID_PAUSE_RESUME, sensitive
	);
	s = win->paused ? "Resume" : "Pause";
	icon_data = win->paused ?
		(guint8 **)icon_resume_20x20_xpm :
		(guint8 **)icon_pause_20x20_xpm;
	GUIMenuItemSetLabel(win->pause_resume_mi, s);
	GUIMenuItemSetPixmap(win->pause_resume_mi, icon_data);
	GUIButtonPixmapUpdate(win->pause_resume_btn, icon_data, s);
	opid = avs_win_ops_list_match_by_id(
		win->opids_list,
		AVS_WIN_OPID_PAUSE_RESUME
	);
	if(opid != NULL)
		GUIButtonLabelUnderline(win->pause_resume_btn, opid->accel_key);
	GUIMenuItemSetLabel(win->scan_list_menu_pause_resume_mi, s);
	GUIMenuItemSetPixmap(win->scan_list_menu_pause_resume_mi, icon_data);
	ToolBarUpdateItem(
		tool_bar, AVS_WIN_OPID_PAUSE_RESUME,
		s, icon_data, NULL
	);


	/* Scan Options */
	sensitive = TRUE;
/*	sensitive = (scanning || db_updating) ? FALSE : TRUE; */
	GTK_WIDGET_SET_SENSITIVE(win->location_mi, sensitive);
	GTK_WIDGET_SET_SENSITIVE(win->scan_location_entry, sensitive);
	GTK_WIDGET_SET_SENSITIVE(win->scan_location_browse_btn, sensitive);
	ToolBarSetItemSensitive(
		tool_bar, AVS_WIN_OPID_LOCATION, sensitive
	);
	GTK_WIDGET_SET_SENSITIVE(win->recursive_micheck, sensitive);
	GTK_WIDGET_SET_SENSITIVE(win->scan_recursive_check, sensitive);
	ToolBarSetItemSensitive(
		tool_bar, AVS_WIN_OPID_RECURSIVE, sensitive
	);
	GTK_WIDGET_SET_SENSITIVE(win->executables_only_micheck, sensitive);
	GTK_WIDGET_SET_SENSITIVE(win->scan_executables_only_check, sensitive);
	ToolBarSetItemSensitive(
		tool_bar, AVS_WIN_OPID_EXECUTABLES_ONLY, sensitive
	);
	GTK_WIDGET_SET_SENSITIVE(win->ignore_links_micheck, sensitive);
	GTK_WIDGET_SET_SENSITIVE(win->scan_ignore_links_check, sensitive);
	ToolBarSetItemSensitive(
		tool_bar, AVS_WIN_OPID_IGNORE_LINKS, sensitive
	);

	sensitive = scanning ? FALSE : TRUE;
	GTK_WIDGET_SET_SENSITIVE(win->scanner_options_mi, sensitive);
	GTK_WIDGET_SET_SENSITIVE(win->scan_scanner_options_btn, sensitive);
	ToolBarSetItemSensitive(
		tool_bar, AVS_WIN_OPID_SCANNER_SETTINGS, sensitive
	);

	/* Results Recycle & Move */
	sensitive = ((page_num == AVS_WIN_PAGE_RESULTS) &&
		(last_selected_results_item > -1)) ? TRUE : FALSE;
	GTK_WIDGET_SET_SENSITIVE(win->results_recycle_mi, sensitive);
	GTK_WIDGET_SET_SENSITIVE(win->results_list_menu_recycle_mi, sensitive);
	GTK_WIDGET_SET_SENSITIVE(win->results_recycle_btn, sensitive);
	ToolBarSetItemSensitive(
		tool_bar, AVS_WIN_OPID_RESULTS_RECYCLE, sensitive
	);
	GTK_WIDGET_SET_SENSITIVE(win->results_move_mi, sensitive);
	GTK_WIDGET_SET_SENSITIVE(win->results_list_menu_move_mi, sensitive);
	GTK_WIDGET_SET_SENSITIVE(win->results_move_btn, sensitive);
	ToolBarSetItemSensitive(
		tool_bar, AVS_WIN_OPID_RESULTS_MOVE, sensitive
	);

	/* Results Open, Save As, and Clear */ 
	sensitive = ((page_num == AVS_WIN_PAGE_RESULTS) &&
		!scanning) ? TRUE : FALSE;
	GTK_WIDGET_SET_SENSITIVE(win->results_report_mi, sensitive);
	GTK_WIDGET_SET_SENSITIVE(win->results_open_mi, sensitive);
	GTK_WIDGET_SET_SENSITIVE(win->results_open_btn, sensitive);
	GTK_WIDGET_SET_SENSITIVE(win->results_save_as_mi, sensitive);
	GTK_WIDGET_SET_SENSITIVE(win->results_save_as_btn, sensitive);
	GTK_WIDGET_SET_SENSITIVE(win->results_clear_mi, sensitive);
	GTK_WIDGET_SET_SENSITIVE(win->results_clear_btn, sensitive);
	GTK_WIDGET_SET_SENSITIVE(win->results_report_btn, sensitive);
	ToolBarSetItemSensitive(
		tool_bar, AVS_WIN_OPID_RESULTS_REPORT, sensitive
	);
	ToolBarSetItemSensitive(
		tool_bar, AVS_WIN_OPID_RESULTS_OPEN, sensitive
	);
	ToolBarSetItemSensitive(
		tool_bar, AVS_WIN_OPID_RESULTS_SAVE_AS, sensitive
	);
	ToolBarSetItemSensitive(
		tool_bar, AVS_WIN_OPID_RESULTS_CLEAR, sensitive
	);


	/* Virus Database Refresh */
	sensitive = ((page_num != AVS_WIN_PAGE_DB) ||
		scanning || db_updating) ? FALSE : TRUE;
	GTK_WIDGET_SET_SENSITIVE(win->db_refresh_mi, sensitive);
	GTK_WIDGET_SET_SENSITIVE(win->db_refresh_btn, sensitive);
	ToolBarSetItemSensitive(
		tool_bar, AVS_WIN_OPID_DB_REFRESH, sensitive
	);

	/* Virus Database Location */
	sensitive = (scanning || db_updating) ? FALSE : TRUE;
	GTK_WIDGET_SET_SENSITIVE(win->db_location_mi, sensitive);
	GTK_WIDGET_SET_SENSITIVE(win->scan_db_location_btn, sensitive);
	GTK_WIDGET_SET_SENSITIVE(win->db_location_btn, sensitive);
	ToolBarSetItemSensitive(
		tool_bar, AVS_WIN_OPID_DB_LOCATION, sensitive
	);

	/* Virus Database Update Net */
	sensitive = (scanning || db_updating) ? FALSE : TRUE;
	GTK_WIDGET_SET_SENSITIVE(win->db_update_net_mi, sensitive);
	GTK_WIDGET_SET_SENSITIVE(win->db_update_net_btn, sensitive);
	ToolBarSetItemSensitive(
		tool_bar, AVS_WIN_OPID_DB_UPDATE_NET, sensitive
	);


	/* View Scan Page */
	toggled = (page_num == AVS_WIN_PAGE_SCAN) ? TRUE : FALSE;
	GUIMenuItemSetCheck(
		win->view_scan_page_micheck,
		toggled, TRUE
	);
	ToolBarSetItemToggle(
		tool_bar, AVS_WIN_OPID_VIEW_SCAN_PAGE, toggled
	);

	/* View Results Page */
	toggled = (page_num == AVS_WIN_PAGE_RESULTS) ? TRUE : FALSE;
	GUIMenuItemSetCheck(
		win->view_results_page_micheck,
		toggled, TRUE
	);
	ToolBarSetItemToggle(
		tool_bar, AVS_WIN_OPID_VIEW_RESULTS_PAGE, toggled
	);

	/* View Virus Database Page */
	toggled = (page_num == AVS_WIN_PAGE_DB) ? TRUE : FALSE;
	GUIMenuItemSetCheck(
		win->view_db_page_micheck,
		toggled, TRUE
	);
	ToolBarSetItemToggle(
		tool_bar, AVS_WIN_OPID_VIEW_DB_PAGE, toggled
	);

	/* View Tool Bar */
	toggled = win->tool_bar_map_state;
	GUIMenuItemSetCheck(
		win->view_tool_bar_micheck,
		toggled, TRUE
	);
	ToolBarSetItemToggle(
		tool_bar, AVS_WIN_OPID_VIEW_TOOLBAR, toggled
	);

	/* View Status Bar */
	toggled = win->status_bar_map_state;
	GUIMenuItemSetCheck(
		win->view_status_bar_micheck,
		toggled, TRUE
	);
	ToolBarSetItemToggle(
		tool_bar, AVS_WIN_OPID_VIEW_STATUSBAR, toggled
	);


	win->freeze_count--;
}

/*
 *	Sets the Win as busy or ready.
 */
void avs_win_set_busy(
	AVSWin *win,
	const gboolean busy
)
{
	GdkCursor *cur;
	GdkWindow *window;
	GtkWidget *w;
	AVSCore *core;

	if(win == NULL)
		return;

	w = win->toplevel;
	core = win->core;
	if(core == NULL)
		return;

	if(busy)
	{
		win->busy_count++;
		if(win->busy_count > 1)
			return;
		cur = core->busy_cur;
	}
	else
	{
		win->busy_count--;
		if(win->busy_count < 0)
			win->busy_count = 0;
		if(win->busy_count > 0)
			return;
		cur = (win->passive_busy_count > 0) ?
			core->passive_busy_cur : NULL;
	}

	window = w->window;
	if(window != NULL)
	{
		gdk_window_set_cursor(window, cur);
		gdk_flush();
	}
}

/*
 *	Sets the Win as passive busy or ready.
 */
void avs_win_set_passive_busy(
	AVSWin *win,
	const gboolean busy
)
{
	GdkCursor *cur;
	GtkWidget *w;
	AVSCore *core;

	if(win == NULL)
		return;

	w = win->toplevel;
	core = AVS_CORE(win->core);

	if(busy)
	{
		win->passive_busy_count++;
		if((win->passive_busy_count > 1) || (win->busy_count > 0))
			return;

		cur = core->passive_busy_cur;
	}
	else
	{
		win->passive_busy_count--;
		if(win->passive_busy_count < 0)
			win->passive_busy_count = 0;
		if((win->passive_busy_count > 0) || (win->busy_count > 0))
			return;

		cur = NULL;
	}
	if(w->window != NULL)
	{
		gdk_window_set_cursor(
			w->window,
			cur
		);
		gdk_flush();
	}
}

/*                          
 *	Checks if the Win's Tool Bar is shown.      
 */
gboolean avs_win_tool_bar_is_shown(AVSWin *win)
{
	if(win == NULL)
		return(FALSE);

	return(win->tool_bar_map_state);
}

/*
 *	Show/hide the Win's Tool Bar.
 */
void avs_win_show_tool_bar(
	AVSWin *win,
	const gboolean show
)
{
	GtkWidget *w;

	if(win == NULL)
		return;

	win->freeze_count++;

	w = win->tool_bar_handle;

	if(win->tool_bar_map_state && !show)
	{
		gtk_widget_hide(w);
		win->tool_bar_map_state = FALSE;
	}
	else if(!win->tool_bar_map_state && show)
	{
		gtk_widget_show(w);
		win->tool_bar_map_state = TRUE;
	}

	avs_win_update_display(win);

	win->freeze_count--;
}

/*
 *	Recreates the Win's Tool Bar Items.
 */
void avs_win_update_tool_bar(AVSWin *win)
{
	if(win == NULL)
		return;

	/* Recreate the Tool Bar and the Tool Bar Items */
	avs_win_build_tool_bar(win);
}

/*                          
 *	Checks if the Win's Status Bar is shown.
 */
gboolean avs_win_status_bar_is_shown(AVSWin *win)
{
	if(win == NULL)
		return(FALSE);

	return(win->status_bar_map_state);
}

/*
 *	Show/hide the Win's Status Bar.
 */
void avs_win_show_status_bar(
	AVSWin *win,
	const gboolean show
)
{
	GtkWidget *w;

	if(win == NULL)
		return;

	win->freeze_count++;

	w = win->status_bar;

	if(win->status_bar_map_state && !show)
	{
		gtk_widget_hide(w);
		win->status_bar_map_state = FALSE;
	}
	else if(!win->status_bar_map_state&& show)
	{
		gtk_widget_show(w);
		win->status_bar_map_state = TRUE;
	}

	avs_win_update_display(win);

	win->freeze_count--;
}

/*
 *	Sets the status bar progress.
 */
void avs_win_set_status_progress(
	AVSWin *win,
	const gfloat v,
	const gboolean allow_gtk_iteration
)
{
	gfloat val = v;
	GtkWidget *w;
	GtkProgress *progress;
	GtkProgressBar *progress_bar;

	if(win == NULL)
		return;

	w = win->status_bar_progress;
	progress = GTK_PROGRESS(w);
	progress_bar = GTK_PROGRESS_BAR(w);

	/* Do activity? */
	if(val < 0.0f)
	{
		GtkAdjustment *adj = progress->adjustment;

		/* Get the new value based on the unknown position */
		val = gtk_progress_get_value(progress) + 1.0f;
		if(val > adj->upper)
			val = adj->lower;

		/* Update progress bar for `activity mode' (unknown limit)
		 * and set new value            
		 */
		gtk_progress_set_activity_mode(progress, TRUE);
		gtk_progress_set_show_text(progress, FALSE);
		gtk_progress_set_value(progress, val);

		/* Reset last progress position to -1.0, indicating that
		 * there is no known last position since we are doing
		 * activity mode
		 */
		win->status_bar_progress_last = -1.0f;
	}
	else
	{
		/* Clip value to 0.0, 1.0 */
		if(val > 1.0f)
			val = 1.0f;

		/* Reset last progress position if it is greater than the
		 * given value, implying the progress has warped back to
		 * the beginning
		 */
		if(win->status_bar_progress_last > val)
			win->status_bar_progress_last = -1.0f;

		/* Has value not sufficiently changed? (less than 0.009
		 * change in value)
		 */
		if((win->status_bar_progress_last > 0.0f) &&
		   ((val - win->status_bar_progress_last) < 0.009f)
		)
			return;

		/* Update progress */ 
		gtk_progress_set_activity_mode(progress, FALSE);
		gtk_progress_set_format_string(
			progress,
			"%p%%"
		);
		/* Display text only if value is positive */
		gtk_progress_set_show_text(
			progress,
			(val > 0.0f) ? TRUE : FALSE 
		);
		gtk_progress_bar_update(progress_bar, val); 

		/* Record the last progress value */
		win->status_bar_progress_last = val;
	}

	if(allow_gtk_iteration)
	{
		while(gtk_events_pending() > 0)
			gtk_main_iteration();
	}
}

/*
 *	Sets the status bar message.
 */
void avs_win_set_status_message(
	AVSWin *win,
	const gchar *msg,
	const gboolean allow_gtk_iteration
)
{
	GtkWidget *w;

	if(win == NULL)
		return;

	w = win->status_bar_label;

	gtk_label_set_text(
		GTK_LABEL(w),
		(msg != NULL) ? msg : ""
	);

	if(allow_gtk_iteration)
	{
		while(gtk_events_pending() > 0)
			gtk_main_iteration();
	}
}

/*
 *	Set's the status bar time.
 */
void avs_win_set_status_time(
	AVSWin *win,
	const gulong t,
	const gboolean allow_gtk_iteration
)
{
	gchar *s;

	if(win == NULL)
		return;

	/* Longer than 1 hour? */
	if(t >= (60l * 60l))
		s = g_strdup_printf(
			"%i:%.2i:%.2i",
			(gint)(t / (60l * 60l)),
			(gint)((t / 60) % 60l),
			(gint)(t % 60l)
		);
	else	/* Less than 1 hour */
		s = g_strdup_printf(
			"%i:%.2i",
			(gint)(t / 60l),
			(gint)(t % 60l)
		);
	gtk_label_set_text(GTK_LABEL(win->status_bar_time_label), s);
	g_free(s);

	/* Record the status time */
	win->status_time = t;

	if(allow_gtk_iteration)
	{
		while(gtk_events_pending() > 0)
			gtk_main_iteration();
	}
}

/*
 *	Checks if the Win is mapped.
 */
gboolean avs_win_is_mapped(AVSWin *win)
{
	if(win == NULL)
		return(FALSE);

	return(GTK_WIDGET_MAPPED(win->toplevel));
}

/*
 *	Maps the Win.
 */
void avs_win_map(AVSWin *win)
{
	if(win == NULL)
		return;

	gtk_widget_show_raise(win->toplevel);
}

/*
 *	Unmaps the Win.
 */
void avs_win_unmap(AVSWin *win)
{
	if(win == NULL)
		return;

	gtk_widget_hide(win->toplevel);
}

/*
 *	Deletes the Win.
 */
void avs_win_delete(AVSWin *win)
{
	if(win == NULL)
		return;

	/* Interrupt all child processes, close and remove the stdout
	 * and stderr files, and remove all timeouts
	 */
	avs_win_stop_scan_process(win);
	avs_win_stop_db_update_net(win);

	/* Remove all the idle and timeout callbacks */
	win->db_update_display_idleid = GTK_IDLE_REMOVE(win->db_update_display_idleid);


	avs_win_unmap(win);

	win->freeze_count++;

	/* Destroy all Right-Click GtkMenus */
	gtk_widget_destroy(win->scan_list_menu);
	gtk_widget_destroy(win->results_list_menu);

	/* Destroy the rest of the GtkWidgets */
	gtk_widget_destroy(win->toplevel);

	gtk_accel_group_unref(win->accelgrp);

	/* Delete the AVSWinOps List */
	avs_win_ops_list_delete(win->opids_list);

	/* Unref the GtkStyles */
	GTK_STYLE_UNREF(win->infected_style);

	/* Delete the scan context, any members should have been
	 * deleted in avs_win_stop_scan_process()
	 */
	g_free(win->scan_context);

	win->freeze_count--;

	g_free(win);
}
