/*
                              AVScan Window
 */

#ifndef AVS_WIN_H
#define AVS_WIN_H

#include <stdio.h>
#include <gtk/gtk.h>
#include "tool_bar.h"
#include "obj.h"


typedef struct _AVSWinOp		AVSWinOp;
#define AVS_WIN_OP(p)			((AVSWinOp *)(p))

typedef struct _AVSWin			AVSWin;
#define AVS_WIN(p)			((AVSWin *)(p))

typedef struct _AVSWinScanContext	AVSWinScanContext;
#define AVS_WIN_SCAN_CONTEXT(p)		((AVSWinScanContext *)(p))


/*
 *	Operation IDs:
 */
typedef enum {

	AVS_WIN_OPID_SEPARATOR		= 10,

	AVS_WIN_OPID_CLOSE		= 100,
	AVS_WIN_OPID_EXIT		= 101,

	AVS_WIN_OPID_ADD		= 200,
	AVS_WIN_OPID_EDIT		= 201,
	AVS_WIN_OPID_REMOVE		= 202,

	AVS_WIN_OPID_NEW		= 300,
	AVS_WIN_OPID_CUT		= 310,
	AVS_WIN_OPID_COPY		= 311,
	AVS_WIN_OPID_PASTE		= 312,
	AVS_WIN_OPID_DELETE		= 320,
	AVS_WIN_OPID_SELECT_ALL		= 330,
	AVS_WIN_OPID_UNSELECT_ALL	= 331,
	AVS_WIN_OPID_INVERT_SELECTION	= 332,
	AVS_WIN_OPID_PROPERTIES		= 340,

	AVS_WIN_OPID_START		= 400,
	AVS_WIN_OPID_STOP		= 401,
	AVS_WIN_OPID_PAUSE_RESUME	= 402,
	AVS_WIN_OPID_LOCATION		= 410,
	AVS_WIN_OPID_RECURSIVE		= 411,
	AVS_WIN_OPID_EXECUTABLES_ONLY	= 412,
	AVS_WIN_OPID_IGNORE_LINKS	= 413,
	AVS_WIN_OPID_SCANNER_SETTINGS	= 490,

	AVS_WIN_OPID_RESULTS_RECYCLE	= 500,
	AVS_WIN_OPID_RESULTS_MOVE	= 501,
	AVS_WIN_OPID_RESULTS_SELECT_ALL	= 510,
	AVS_WIN_OPID_RESULTS_UNSELECT_ALL	\
					= 511,
	AVS_WIN_OPID_RESULTS_INVERT_SELECTION	\
					= 512,
	AVS_WIN_OPID_RESULTS_REPORT	= 520,
	AVS_WIN_OPID_RESULTS_OPEN	= 530,
	AVS_WIN_OPID_RESULTS_SAVE_AS	= 531,
	AVS_WIN_OPID_RESULTS_CLEAR	= 540,

	AVS_WIN_OPID_DB_REFRESH		= 600,
	AVS_WIN_OPID_DB_LOCATION	= 620,
	AVS_WIN_OPID_DB_LIST_PATTERNS	= 630,
	AVS_WIN_OPID_DB_UPDATE_NET	= 670,

	AVS_WIN_OPID_VIEW_SCAN_PAGE	= 710,
	AVS_WIN_OPID_VIEW_RESULTS_PAGE	= 711,
	AVS_WIN_OPID_VIEW_DB_PAGE	= 712,
	AVS_WIN_OPID_VIEW_TOOLBAR	= 720,
	AVS_WIN_OPID_VIEW_STATUSBAR	= 721,

	AVS_WIN_OPID_HELP_CONTENTS	= 1000,
	AVS_WIN_OPID_HELP_SCANNING	= 1001,
	AVS_WIN_OPID_HELP_UPDATING	= 1002,
	AVS_WIN_OPID_ABOUT		= 1100

} AVSWinOpID;


/*
 *	Operation:
 */
struct _AVSWinOp {

	AVSWinOpID	id;
	AVSWin	*win;        

	gboolean	allow_multiple;
	ToolBarItemType	item_type;

	gchar		*button_name,
			*menu_name,
			*tooltip;

	guint		accel_key,
			accel_mods;

	guint8		**button_icon_data,	/* Shared */
			**menu_icon_data;	/* Shared */

	void (*func_cb)(
		GtkWidget *,			/* GtkWidget */
		gpointer			/* Win */
	);
	gint (*enter_func_cb)(
		GtkWidget *,			/* GtkWidget */
		GdkEventCrossing *,		/* GdkEventCrossing */
		gpointer			/* OPID */
	);
	gint (*leave_func_cb)(
		GtkWidget *,			/* GtkWidget */
		GdkEventCrossing *,		/* GdkEventCrossing */
		gpointer			/* OPID */             
	);

};


/*
 *	Page Indicies:
 */
typedef enum {
	AVS_WIN_PAGE_SCAN,
	AVS_WIN_PAGE_RESULTS,
	AVS_WIN_PAGE_DB
} AVSWinPageIndex;


/*
 *	AVScan Window:
 */
struct _AVSWin {

	GtkWidget	*toplevel;
	GtkAccelGroup	*accelgrp;
	gboolean	processing,
			paused;
	gint		busy_count,
			passive_busy_count,
			freeze_count,
			stop_count;
	gpointer	core;

	GList		*opids_list;

	GtkWidget	*main_vbox;

	gboolean	menubar_map_state;
	GtkWidget	*menubar_handle,
			*menubar;

	GtkWidget	*scan_mlabel,
			*new_mi,
			*location_mi,
			*recursive_micheck,
			*executables_only_micheck,
			*ignore_links_micheck,
			*scanner_options_mi,
			*start_mi,
			*stop_mi,
			*pause_resume_mi,
			*close_mi,
			*edit_mlabel,
			*cut_mi,
			*copy_mi,
			*paste_mi,
			*delete_mi,
			*properties_mi,
			*view_mlabel,
			*view_scan_page_micheck,
			*view_results_page_micheck,
			*view_db_page_micheck,
			*view_tool_bar_micheck,
			*view_status_bar_micheck,
			*results_mlabel,
			*results_recycle_mi,
			*results_move_mi,
			*results_select_all_mi,
			*results_unselect_all_mi,
			*results_invert_selection_mi,
			*results_report_mi,
			*results_open_mi,
			*results_save_as_mi,
			*results_clear_mi,
			*db_mlabel,
			*db_refresh_mi,
			*db_location_mi,
			*db_update_net_mi,
			*help_mlabel,
			*help_contents_mi,
			*about_mi;

	gboolean	tool_bar_map_state;
	GtkWidget	*tool_bar_handle;
	GtkWidget	*tool_bar;

	AVSWinPageIndex	current_page_index;

	GtkWidget	*lists_vbox,
			*notebook,
			*scan_clist,
			*scan_add_btn,
			*scan_edit_btn,
			*scan_remove_btn,
			*scan_location_entry,
			*scan_location_browse_btn,
			*scan_recursive_check,
			*scan_executables_only_check,
			*scan_ignore_links_check,
			*scan_scanner_options_btn,
			*scan_db_location_btn,
			*results_location_label,
			*results_status_label,
			*results_report_btn,
			*results_clist,
			*results_recycle_btn,
			*results_move_btn,
			*results_clear_btn,
			*results_open_btn,
			*results_save_as_btn,
			*db_name_label,
			*db_location_entry,
			*db_updated_label,	/* Last updated */
			*db_version_label,
			*db_build_time_label,
			*db_patterns_label,
			*db_size_label,
			*db_signature_entry,
			*db_md5_entry,
			*db_builder_entry,
			*db_functionality_level_label,
			*db_refresh_btn,
			*db_location_btn,
			*db_update_net_btn,
			*anim_icon,
			*start_btn,
			*stop_btn,
			*pause_resume_btn,
			*close_btn;

	/* Status Bar */
	gboolean	status_bar_map_state;
	GtkWidget	*status_bar;
	GtkWidget	*status_bar_progress,
			*status_bar_label,
			*status_bar_time_label;
	gfloat		status_bar_progress_last;
	gulong		status_time;

	/* Scan List right-click menu */
	GtkWidget	*scan_list_menu,
			*scan_list_menu_start_mi,
			*scan_list_menu_stop_mi,
			*scan_list_menu_pause_resume_mi,
			*scan_list_menu_new_mi,
			*scan_list_menu_cut_mi,
			*scan_list_menu_copy_mi,
			*scan_list_menu_paste_mi,
			*scan_list_menu_delete_mi,
			*scan_list_menu_properties_mi;

	/* Results List right-click menu */
	GtkWidget	*results_list_menu,
			*results_list_menu_recycle_mi,
			*results_list_menu_move_mi,
			*results_list_menu_select_all_mi,
			*results_list_menu_unselect_all_mi,
			*results_list_menu_invert_selection_mi;

	/* Styles */
	GtkStyle	*infected_style;


	/* Scan Context */
	AVSWinScanContext	*scan_context;

	/* Update Virus Database display */
	guint		db_update_display_idleid;

	/* Update Virus Database Process */
	guint		db_update_net_toid;
	gint		db_update_net_pid;
	gulong		db_update_net_start_time;

	/* Child process' output streams and buffers */
	FILE		*cstdout,
			*cstderr;
	gchar		*cstdout_line_buf,
			*cstderr_line_buf,
			*cstdout_path,
			*cstderr_path;
};


/*
 *	Scan Context:
 */
struct _AVSWinScanContext {

	guint		toid;			/* Scan monitoring timeout
						 * callback */
	gint		pid;			/* Scan process ID */
	gulong		start_time,		/* In seconds since EPOCH */
			duration;		/* In seconds, set after scan
						 * is complete or interrupted */
	gulong		nscanned,
			ninfected,
			nproblems,
			total;
	gulong		blocks_scanned;		/* 1 block = 1024 bytes */

};


/* Utilities */
extern gchar *avs_win_strdup_shorten(
        const gchar *s,
        const gint max_characters
);

/* Utilities */
extern const gchar *avs_win_current_scan_location(AVSWin *win);
extern void avs_win_set_scan_location(
	AVSWin *win,
	const gchar *location
);
extern GList *avs_win_current_paths_list(AVSWin *win);
extern AVSScanSettings avs_win_current_scan_settings(AVSWin *win);
extern void avs_win_set_scan_settings(
	AVSWin *win,
	const AVSScanSettings options
); 
extern void avs_win_switch_page(
	AVSWin *win,
	const AVSWinPageIndex page_num
);

/* Scan List & Settings Update */
extern void avs_win_scan_list_open_file(
	AVSWin *win,
	const gchar *path
);
extern void avs_win_scan_settings_update(
	AVSWin *win,
	GList *objs_list
);
extern void avs_win_set_results_location(
	AVSWin *win,
	GList *paths_list
);
extern void avs_win_set_results_status(
	AVSWin *win,
	const gchar *s
);

/* Operation IDs */
extern AVSWinOp *avs_win_op_new(
	AVSWinOpID id,
	AVSWin *win
);
extern void avs_win_op_delete(AVSWinOp *opid);
extern GList *avs_win_ops_list_new(AVSWin *win);
extern void avs_win_ops_list_delete(GList *glist);
extern AVSWinOp *avs_win_ops_list_match_by_id(
	GList *ops_list,
	const AVSWinOpID id
);

/* AVScan Window */
extern AVSWin *avs_win_new(gpointer core_ptr);
extern void avs_win_sync_to_cfg_list(AVSWin *win);
extern void avs_win_update_title(AVSWin *win);
extern void avs_win_update_display(AVSWin *win);
extern void avs_win_set_busy(AVSWin *win, const gboolean busy);
extern void avs_win_set_passive_busy(AVSWin *win, const gboolean busy);
extern gboolean avs_win_tool_bar_is_shown(AVSWin *win);
extern void avs_win_show_tool_bar(AVSWin *win, const gboolean show);
extern void avs_win_update_tool_bar(AVSWin *win);
extern gboolean avs_win_status_bar_is_shown(AVSWin *win);
extern void avs_win_show_status_bar(AVSWin *win, const gboolean show);
extern void avs_win_set_status_progress(
	AVSWin *win,
	const gfloat v,
	const gboolean allow_gtk_iteration
);
extern void avs_win_set_status_message(
	AVSWin *win,
	const gchar *msg,
	const gboolean allow_gtk_iteration
);
extern void avs_win_set_status_time(
	AVSWin *win,
	const gulong t,
	const gboolean allow_gtk_iteration
);
extern gboolean avs_win_is_mapped(AVSWin *win);
extern void avs_win_map(AVSWin *win);
extern void avs_win_unmap(AVSWin *win);
extern void avs_win_delete(AVSWin *win);


#endif	/* AVS_WIN_H */
