#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

#include "../include/disk.h"

#include "obj.h"
#include "avs_win.h"    
#include "avs_win_process.h"
#include "avs_win_cb.h"
#include "avs_win_dnd.h"
#include "avs_win_list.h"
#include "avs_win_db.h"
#include "avs_win_op_cb.h"
#include "avs_core.h"


void avs_win_scan_list_item_destroy_cb(gpointer data);
void avs_win_results_list_item_destroy_cb(gpointer data);

gint avs_win_event_cb(
	GtkWidget *widget, GdkEvent *event, gpointer data
);

gint avs_win_op_enter_event_cb(
	GtkWidget *widget, GdkEventCrossing *crossing, gpointer data
);
gint avs_win_op_leave_event_cb(
	GtkWidget *widget, GdkEventCrossing *crossing, gpointer data
);

void avs_win_tool_bar_item_cb(
	ToolBarItem *item, gint id, gpointer data
);
void avs_win_tool_bar_item_enter_cb(
	ToolBarItem *item, gint id, gpointer data
);
void avs_win_tool_bar_item_leave_cb(
	ToolBarItem *item, gint id, gpointer data
);

void avs_win_notebook_switch_page_cb(
	GtkNotebook *notebook, GtkNotebookPage *page, guint page_num,
	gpointer data
);

void avs_win_location_entry_changed_cb(GtkWidget *widget, gpointer data);
gint avs_win_entry_complete_path_key_event_cb(
	GtkWidget *widget, GdkEventKey *key, gpointer data
);

gint avs_win_list_event_cb(GtkWidget *widget, GdkEvent *event, gpointer data);
void WinListResizeColumnCB(
	GtkCList *clist, gint column, gint width, gpointer data
);
void avs_win_list_click_column_cb(
	GtkCList *clist, gint column, gpointer data             
);
void avs_win_list_select_row_cb(
	GtkCList *clist, gint row, gint column, GdkEvent *event,
	gpointer data
);
void avs_win_list_unselect_row_cb(
	GtkCList *clist, gint row, gint column, GdkEvent *event,
	gpointer data
);

void avs_win_tree_select_row_cb(
	GtkCTree *ctree, GtkCTreeNode *node, gint column,
	gpointer data
);
void avs_win_tree_unselect_row_cb(
	GtkCTree *ctree, GtkCTreeNode *node, gint column,
	gpointer data
);
void avs_win_tree_expand_cb(
	GtkCTree *ctree, GtkCTreeNode *node, gpointer data
);
void avs_win_tree_collapse_cb(
	GtkCTree *ctree, GtkCTreeNode *node, gpointer data
);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : TRUE)


/*
 *	Scan List item destroy signal callback.
 */
void avs_win_scan_list_item_destroy_cb(gpointer data)
{
	ObjDelete(OBJ(data));
}

/*
 *	Results List item destroy signal callback.
 */
void avs_win_results_list_item_destroy_cb(gpointer data)
{
	ObjDelete(OBJ(data));
}

/*
 *	Win toplevel GtkWindow event signal callback.
 */
gint avs_win_event_cb(
	GtkWidget *widget, GdkEvent *event, gpointer data
)
{
	gint status = FALSE;
	gboolean press;
	guint keyval, state;
	GdkEventKey *key;
	AVSWin *win = AVS_WIN(data);
	if((widget == NULL) || (event == NULL) || (win == NULL))
		return(status);

	switch((gint)event->type)
	{
	  case GDK_DELETE:
		/* If there a scan, database load, or virus database update
		 * in progress then call the stop callback instead of
		 * the close callback
		 */
		if(avs_win_scan_process_is_running(win) ||
		   (win->db_update_display_idleid != 0) ||
		   avs_win_db_update_net_process_is_running(win)
		)
			WinStopCB(win->stop_mi, win);
		else
			WinCloseCB(win->close_mi, win);
		status = TRUE;
		break;

	  case GDK_KEY_PRESS:
	  case GDK_KEY_RELEASE:
		key = (GdkEventKey *)event;
		press = (event->type == GDK_KEY_PRESS) ? TRUE : FALSE;
		keyval = key->keyval;
		state = key->state;
#define DO_STOP_KEY_SIGNAL_EMIT {		\
 gtk_signal_emit_stop_by_name(			\
  GTK_OBJECT(widget),				\
  press ?                                       \
   "key_press_event" : "key_release_event"      \
 );						\
}
		switch(keyval)
		{
		  case GDK_Escape:
			if(press)
				WinStopCB(win->stop_mi, win);
			DO_STOP_KEY_SIGNAL_EMIT
			status = TRUE;
			break;

		}
#undef DO_STOP_KEY_SIGNAL_EMIT
		break;
	}

	return(status);
}

/*
 *	Win OPID "enter_notify_event" signal callback.
 */
gint avs_win_op_enter_event_cb(
	GtkWidget *widget, GdkEventCrossing *crossing, gpointer data
)
{
	AVSWin *win;
	AVSWinOp *opid = AVS_WIN_OP(data);
	if(opid == NULL)
		return(FALSE);

	win = opid->win;
	if(win == NULL)
		return(FALSE);

	if(win->freeze_count > 0)
		return(FALSE);

	avs_win_set_status_message(win, opid->tooltip, FALSE);

	return(TRUE);
}

/*
 *	Win OPID "leave_notify_event" signal callback.
 */
gint avs_win_op_leave_event_cb(
	GtkWidget *widget, GdkEventCrossing *crossing, gpointer data
)
{
	AVSWin *win;
	AVSWinOp *opid = AVS_WIN_OP(data);
	if(opid == NULL)
		return(FALSE);

	win = opid->win;
	if(win == NULL)
		return(FALSE);

	if(win->freeze_count > 0)
		return(FALSE);

	avs_win_set_status_message(win, NULL, FALSE);

	return(TRUE);
}


/*
 *	Win Tool Bar Item callback.
 */
void avs_win_tool_bar_item_cb(
	ToolBarItem *item, gint id, gpointer data
)
{
	AVSWin *win;
	AVSWinOp *opid = AVS_WIN_OP(data);
	if(opid == NULL)
		return;

	win = opid->win;
	if(win == NULL)
		return;

	if(win->freeze_count > 0)
		return;

	if(opid->func_cb != NULL)
		opid->func_cb(
			ToolBarItemGetWidget(item),	/* GtkWidget */
			win				/* Win */
		);
}

/*
 *	Win Tool Bar Item enter callback.
 */
void avs_win_tool_bar_item_enter_cb(
	ToolBarItem *item, gint id, gpointer data
)
{
	AVSWinOp *opid = AVS_WIN_OP(data);
	if((item == NULL) || (opid == NULL))
		return;

	avs_win_op_enter_event_cb(
		ToolBarItemGetWidget(item),
		NULL,
		opid
	);
}

/*
 *	Win Tool Bar Item leave callback.
 */
void avs_win_tool_bar_item_leave_cb(
	ToolBarItem *item, gint id, gpointer data
)
{
	AVSWinOp *opid = AVS_WIN_OP(data);
	if((item == NULL) || (opid == NULL))
		return;

	avs_win_op_leave_event_cb(
		ToolBarItemGetWidget(item),
		NULL,
		opid
	);
}


/*
 *	Win notebook "switch_page" signal callback.
 */
void avs_win_notebook_switch_page_cb(
	GtkNotebook *notebook, GtkNotebookPage *page, guint page_num,
	gpointer data
)
{
	AVSWin *win = AVS_WIN(data);
	if(win == NULL)
		return;

	if(win->freeze_count > 0)
		return;

	win->freeze_count++;

	/* Change in page? */
	if(win->current_page_index != (AVSWinPageIndex)page_num)
	{
		/* Record the new page number */
		win->current_page_index = (AVSWinPageIndex)page_num;
		switch(page_num)
		{
		  case AVS_WIN_PAGE_SCAN:
			break;
		  case AVS_WIN_PAGE_RESULTS:
			break;
		  case AVS_WIN_PAGE_DB:
			avs_win_db_queue_update(win);
			break;
		}
		avs_win_update_title(win);
		avs_win_update_display(win);
	}

	win->freeze_count--;
}


/*
 *	Location GtkEntry "changed" signal callback.
 */
void avs_win_location_entry_changed_cb(GtkWidget *widget, gpointer data)
{
	AVSWin *win = AVS_WIN(data);
	if((widget == NULL) || (win == NULL))
		return;

	if(win->freeze_count > 0)
		return;

	win->freeze_count++;

	avs_win_update_display(win);

	win->freeze_count--;
}

/*
 *	Entry complete path "key_press_event" or "key_release_event"
 *	signal callback.
 */
gint avs_win_entry_complete_path_key_event_cb(
	GtkWidget *widget, GdkEventKey *key, gpointer data
)
{
	gint status = FALSE;
	gboolean press;
	GtkEntry *entry = GTK_ENTRY(widget);
	AVSWin *win = AVS_WIN(data);
	if((entry == NULL) || (key == NULL) || (win == NULL))
		return(status);

	press = (key->type == GDK_KEY_PRESS) ? TRUE : FALSE;

#define SIGNAL_EMIT_STOP	{		\
 gtk_signal_emit_stop_by_name(			\
  GTK_OBJECT(widget),				\
  press ?					\
   "key_press_event" : "key_release_event"	\
 );						\
}

	switch(key->keyval)
	{
	  case GDK_Tab:
		if(press)
		{
			gchar *s = STRDUP(gtk_entry_get_text(entry));
			if(s != NULL)
			{
				gint status;

				s = CompletePath(s, &status);
				gtk_entry_set_text(entry, (s != NULL) ? s : "");
				gtk_entry_set_position(entry, -1);
				g_free(s);

				switch(status)
				{
				  case COMPLETE_PATH_NONE:
				  case COMPLETE_PATH_AMBIGUOUS:
					gdk_beep();
					break;
				}
			}
		}
		SIGNAL_EMIT_STOP
		status = TRUE;
		break;
	}

#undef SIGNAL_EMIT_STOP

	return(status);
}


/*
 *	Win list event signal callback.
 */
gint avs_win_list_event_cb(GtkWidget *widget, GdkEvent *event, gpointer data)
{
	gint status = FALSE;
	gboolean press;
	guint keyval, state;
	GdkEventKey *key;
	GdkEventButton *button;
	GdkEventMotion *motion;
	AVSWin *win = AVS_WIN(data);
	if((widget == NULL) || (event == NULL) || (win == NULL))
		return(status);

	switch((gint)event->type)
	{
	  case GDK_KEY_PRESS:
	  case GDK_KEY_RELEASE:
		key = (GdkEventKey *)event;
		press = (event->type == GDK_KEY_PRESS) ? TRUE : FALSE;
		keyval = key->keyval;
		state = key->state;
#define DO_STOP_KEY_SIGNAL_EMIT {		\
 gtk_signal_emit_stop_by_name(			\
  GTK_OBJECT(widget),				\
  press ?                                       \
   "key_press_event" : "key_release_event"      \
 );						\
}
		switch(keyval)
		{
		  case GDK_Insert:
			if(press)
			{
				if(state & GDK_CONTROL_MASK)
					WinCopyCB(win->copy_mi, win);
				else if(state & GDK_SHIFT_MASK)
					WinPasteCB(win->paste_mi, win);
			}
			DO_STOP_KEY_SIGNAL_EMIT
			status = TRUE;
			break;

		  case GDK_Delete:
			if(press)
				WinRemoveCB(win->delete_mi, win);
			DO_STOP_KEY_SIGNAL_EMIT
			status = TRUE;
			break;
		}
#undef DO_STOP_KEY_SIGNAL_EMIT
		break;

	  case GDK_BUTTON_PRESS:
		button = (GdkEventButton *)event;
		switch(button->button)
		{
		  case GDK_BUTTON1:
			break;

		  case GDK_BUTTON2:
			break;

		  case GDK_BUTTON3:
			if(widget == win->scan_clist)
			{
				GtkMenu *menu = GTK_MENU(win->scan_list_menu);
				gtk_menu_popup(
					menu, NULL, NULL,
					NULL, NULL,
					button->button, button->time
				);
 		}
			else if(widget == win->results_clist)
			{
				GtkMenu *menu = GTK_MENU(win->results_list_menu);
				gtk_menu_popup(
					menu, NULL, NULL,
					NULL, NULL,
					button->button, button->time
				);
			}
			status = TRUE;
			break;
		}
		break;

	  case GDK_2BUTTON_PRESS:       /* Double-click */
		button = (GdkEventButton *)event;
		switch(button->button)
		{
		  case GDK_BUTTON1:
			break;
		}
		break;

	  case GDK_BUTTON_RELEASE:
		button = (GdkEventButton *)event;

		break;

	  case GDK_MOTION_NOTIFY:
		motion = (GdkEventMotion *)event;

		break;
	}

	return(status);
}


/*
 *	Win GtkCList "resize_column" signal callback.
 */
void WinListResizeColumnCB(
	GtkCList *clist, gint column, gint width, gpointer data
)
{
	AVSWin *win = AVS_WIN(data);
	if((clist == NULL) || (win == NULL))
		return;

	if(win->freeze_count > 0)
		return;

	if(!GTK_WIDGET_SENSITIVE(GTK_WIDGET(clist)))
		return;


}

/*
 *	Win GtkCList "click_column" signal callback.
 */
void avs_win_list_click_column_cb(
	GtkCList *clist, gint column, gpointer data             
)
{
	AVSWin *win = AVS_WIN(data);
	if((clist == NULL) || (win == NULL))
		return;

	if(win->freeze_count > 0)
		return;

	if(!GTK_WIDGET_SENSITIVE(GTK_WIDGET(clist)))
		return;

	gtk_clist_freeze(clist);

	gtk_clist_set_sort_column(clist, column);
	gtk_clist_set_sort_type(
		clist,
		(clist->sort_type == GTK_SORT_ASCENDING) ?
			GTK_SORT_DESCENDING : GTK_SORT_ASCENDING
	);
	gtk_clist_sort(clist);

	gtk_clist_thaw(clist);
}


/*
 *	Win GtkCList "select_row" signal callback.
 */
void avs_win_list_select_row_cb(
	GtkCList *clist, gint row, gint column, GdkEvent *event,
	gpointer data
)
{
	GtkWidget *widget = (GtkWidget *)clist;
	AVSCore *core;
	AVSWin *win = AVS_WIN(data);
	if((clist == NULL) || (win == NULL))
		return;

	if(win->freeze_count > 0)
		return;

	if(!GTK_WIDGET_SENSITIVE(widget))
		return;

	core = AVS_CORE(win->core);

	/* If the row is not visible then scroll to make it visible */
	if(gtk_clist_row_is_visible(clist, row) !=
		GTK_VISIBILITY_FULL
	)
		gtk_clist_moveto(
			clist,
			row, -1,		/* Row, column */
			0.5f, 0.0f		/* Row, column */
		);

	/* Scan GtkCList */
	if(widget == win->scan_clist)
	{
		const obj_struct *obj = OBJ(
			gtk_clist_get_row_data(clist, row)
		);
		if(obj != NULL)
		{
			gint nobjs;
			gchar *msg;
			GList	*glist,
				*objs_list;
 
			/* Update the scan settings */
			nobjs = 0;
	                objs_list = NULL;
		        for(glist = clist->selection;
	                    glist != NULL;
	                    glist = g_list_next(glist)
	                )
			{
	                        objs_list = g_list_append(
	                                objs_list,
	                                OBJ(gtk_clist_get_row_data(
	                                        clist,
	                                        (gint)glist->data
	                                ))
	                        );
				nobjs++;
			}
	                avs_win_scan_settings_update(win, objs_list);
		        g_list_free(objs_list);

			/* Update status bar message */
			if(nobjs > 1)
				msg = g_strdup_printf(
					"%i scan items selected",
					nobjs
				);
			else if(!STRISEMPTY(obj->name))
				msg = g_strdup_printf(
					"Scan item \"%s\" selected",
					obj->name
				);
			else
				msg = STRDUP("Scan item (Untitled) selected");
			avs_win_set_status_message(win, msg, FALSE);
			g_free(msg);
		}

		GUIDNDSetDragIconFromCListSelection(clist);

		avs_win_update_display(win);

		/* Double click? */
		if(event != NULL)
		{
			if(event->type == GDK_2BUTTON_PRESS)
			{
				GdkEventButton *button = (GdkEventButton *)event;
				switch(button->button)
				{
				  case GDK_BUTTON1:
					/* Start scan */
					WinStartCB(GTK_WIDGET(clist), win);
					break;
				}
			}
		}
	}
	/* Results GtkCList */
	else if(widget == win->results_clist)
	{
		const obj_struct *obj = OBJ(
			gtk_clist_get_row_data(clist, row)
		);
		if((obj != NULL) ? !STRISEMPTY(obj->path) : FALSE)
		{
			struct stat stat_buf;
			gchar *msg;
			const gint nobjs = g_list_length(clist->selection);
			const gchar	*path = obj->path,
					*name = g_basename(path);

			if(nobjs > 1)
				msg = g_strdup_printf(
					"%i objects selected",
					nobjs
				);
			else if(lstat(path, &stat_buf))
				msg = g_strdup_printf(
					"Object \"%s\" selected",
					name
				);
#ifdef S_ISREG
			else if(S_ISREG(stat_buf.st_mode))
#else
			else if(FALSE)
#endif
			{
#ifdef HAVE_LIBENDEAVOUR2
				gchar *size_str = STRDUP(edv_str_size_delim(
					(gulong)stat_buf.st_size
				));
#else
				gchar *size_str = g_strdup_printf(
					"%ld",
					(gulong)stat_buf.st_size
				);
#endif
				msg = g_strdup_printf(
					"File \"%s\" selected (%s byte%s)", 
					name,
					size_str,
					(stat_buf.st_size == 1l) ? "" : "s"
				);
			}
#ifdef S_ISDIR
			else if(S_ISDIR(stat_buf.st_mode))
#else
			else if(FALSE)
#endif
				msg = g_strdup_printf(
					"Directory \"%s\" selected",
					name
				);
#ifdef S_ISLNK
			else if(S_ISLNK(stat_buf.st_mode))
#else
			else if(FALSE)
#endif
				msg = g_strdup_printf(
					"Link \"%s\" selected",
					name
				);
			else
				msg = g_strdup_printf(
					"Object \"%s\" selected",
					name
				);

			avs_win_set_status_message(win, msg, FALSE);
			g_free(msg);
		}

		GUIDNDSetDragIconFromCListSelection(clist);

		avs_win_update_display(win);
	}
}

/*
 *	Win GtkCList "unselect_row" signal callback.
 */
void avs_win_list_unselect_row_cb(
	GtkCList *clist, gint row, gint column, GdkEvent *event,
	gpointer data
)
{
	GtkWidget *widget = (GtkWidget *)clist;
	AVSWin *win = AVS_WIN(data);
	if((widget == NULL) || (win == NULL))
		return;

	if(win->freeze_count > 0)
		return;

	if(!GTK_WIDGET_SENSITIVE(widget))
		return;

	/* Scan GtkCList */
	if(widget == win->scan_clist)
	{
		avs_win_set_status_message(win, NULL, FALSE);
		avs_win_update_display(win);
	}
	/* Results GtkCList */
	else if(widget == win->results_clist)
	{
		avs_win_set_status_message(win, NULL, FALSE);
		avs_win_update_display(win);
	}
}


/*
 *	Win GtkCTree "tree_select_row" signal callback.
 */
void avs_win_tree_select_row_cb(
	GtkCTree *ctree, GtkCTreeNode *node, gint column,
	gpointer data
)
{
	GtkWidget *widget = (GtkWidget *)ctree;
	AVSWin *win = AVS_WIN(data);
	if((widget == NULL) || (win == NULL))
		return;

	if(win->freeze_count > 0)
		return;

	if(!GTK_WIDGET_SENSITIVE(widget))
		return;

	/* If the node is not visible then scroll to make it visible */
	if(gtk_ctree_node_is_visible(ctree, node)
	   != GTK_VISIBILITY_FULL
	)
		gtk_ctree_node_moveto(
			ctree,
			node, -1,
			0.5f, 0.0f			/* Row align, column align */
		);

#if 0
	/* Virus Database Tree */
	if(widget == win->db_ctree)
	{
		gchar *msg;
		win_db_item_struct *d = WIN_DB_ITEM(
			gtk_ctree_node_get_row_data(ctree, node)
		);
		if((d != NULL) ? (d->name != NULL) : FALSE)
		{
			switch(d->type)
			{
			  case WIN_DB_ITEM_TYPE_PATTERN:
				msg = g_strdup_printf(
					"Pattern \"%s\" selected",
					d->name
				);
				break;
			  case WIN_DB_ITEM_TYPE_FOLDER:
				msg = g_strdup_printf(
					"Folder \"%s\" selected",
					d->name
				);
				break;
			  default:
				msg = STRDUP("");
				break;
			}
		}
		else
			msg = STRDUP("");
		avs_win_set_status_message(win, msg, FALSE);
		g_free(msg);

		avs_win_update_display(win);
	}
#endif
}

/*
 *	Win GtkCTree "tree_unselect_row" signal callback.
 */
void avs_win_tree_unselect_row_cb(
	GtkCTree *ctree, GtkCTreeNode *node, gint column,
	gpointer data                                    
)
{
	GtkWidget *widget = (GtkWidget *)ctree;
	AVSWin *win = AVS_WIN(data);
	if((widget == NULL) || (win == NULL))
		return;

	if(win->freeze_count > 0)
		return;

	if(!GTK_WIDGET_SENSITIVE(widget))
		return;

#if 0
	/* Virus Database Tree */
	if(widget == win->db_ctree)
	{
		avs_win_update_display(win);
	}
#endif
}
 
/*
 *	Win GtkCTree "tree_expand" signal callback.
 */
void avs_win_tree_expand_cb(
	GtkCTree *ctree, GtkCTreeNode *node, gpointer data
)
{
	GtkWidget *widget = (GtkWidget *)ctree;
	AVSWin *win = AVS_WIN(data);
	if((ctree == NULL) || (win == NULL))
		return;

	if(win->freeze_count > 0)
		return;

	if(!GTK_WIDGET_SENSITIVE(widget))
		return;
}

/*
 *	Win GtkCTree "tree_collapse" signal callback.
 */
void avs_win_tree_collapse_cb(
	GtkCTree *ctree, GtkCTreeNode *node, gpointer data
)
{ 
	GtkWidget *widget = (GtkWidget *)ctree;
	AVSWin *win = AVS_WIN(data);
	if((ctree == NULL) || (win == NULL))
		return;

	if(win->freeze_count > 0)
		return;

	if(!GTK_WIDGET_SENSITIVE(widget))
		return;
}
