#include <string.h>
#include <time.h>
#include <gtk/gtk.h>
#ifdef HAVE_LIBENDEAVOUR2
# include <endeavour2.h>
#endif

#include "guiutils.h"
#include "cdialog.h"

#include "obj.h"

#include "avs_win.h"
#include "avs_win_process.h"
#include "avs_win_op_cb.h"
#include "avs_win_list.h"
#include "avs_win_dnd.h"
#include "avs_core.h"
#include "config.h"


static void avs_win_dnd_drag_data_received_nexus(
	AVSWin *win,
	GtkWidget *w,
	GdkDragContext *dc,
	const guint info,
	GtkSelectionData *selection_data,
	const gint row, const gint column
);

void avs_win_dnd_anim_icon_drag_data_received_cb(
	GtkWidget *widget, GdkDragContext *dc,
	gint x, gint y,
	GtkSelectionData *selection_data, guint info, guint t,
	gpointer data
);

void avs_win_dnd_list_drag_data_get_cb(
	GtkWidget *widget, GdkDragContext *dc,
	GtkSelectionData *selection_data, guint info, guint t,
	gpointer data                                         
);
void avs_win_dnd_list_drag_data_received_cb(
	GtkWidget *widget, GdkDragContext *dc,
	gint x, gint y,                       
	GtkSelectionData *selection_data, guint info, guint t,
	gpointer data                                         
);
void avs_win_dnd_list_drag_data_delete_cb(
	GtkWidget *widget, GdkDragContext *dc, gpointer data
);

void avs_win_dnd_entry_drag_data_received_cb(
	GtkWidget *widget, GdkDragContext *dc,
	gint x, gint y,
	GtkSelectionData *selection_data, guint info, guint t,
	gpointer data                                         
);


#define ATOI(s)         (((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)         (((s) != NULL) ? atol(s) : 0)
#define ATOF(s)         (((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)       (((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define STRLEN(s)       (((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)   (((s) != NULL) ? (*(s) == '\0') : TRUE)


/*
 *	Drag data received nexus.
 */
static void avs_win_dnd_drag_data_received_nexus(
	AVSWin *win,
	GtkWidget *w,
	GdkDragContext *dc,
	const guint info,
	GtkSelectionData *selection_data,
	const gint row, const gint column
)
{
	GtkWidget *toplevel = win->toplevel;
#ifdef HAVE_LIBENDEAVOUR2
	AVSCore *core = win->core;
	EDVContext *edv_ctx = core->edv_ctx;
#endif

	avs_win_set_busy(win, TRUE);

	/* Animation Icon */
	if(w == win->anim_icon)
	{
		GList *paths_list = NULL;
		GtkWidget *toplevel = win->toplevel;

		/* Is there a scan or update already in progress? */
		if((win->scan_context->toid != 0) ||
		   (win->db_update_net_toid != 0)
		)
		{
#ifdef HAVE_LIBENDEAVOUR2
			edv_play_sound_warning(edv_ctx);
#endif
			CDialogSetTransientFor(toplevel);
			CDialogGetResponse(
				"Scan Failed",
"A scan or virus database update is already in progress.\n\
\n\
You may either wait for the current scan or update to\n\
complete itself or click on the stop button to stop the\n\
current scan or update before starting a new scan.",
				NULL,
				CDIALOG_ICON_WARNING,
				CDIALOG_BTNFLAG_OK,
				CDIALOG_BTNFLAG_OK
			);
			CDialogSetTransientFor(NULL);
			avs_win_set_busy(win, FALSE);
			return;
		}

		/* Format the location string based on the target data
		 * type
		 *
		 * String
		 */
		if((info == DND_INFO_TEXT_PLAIN) ||
		   (info == DND_INFO_TEXT_URI_LIST) ||
		   (info == DND_INFO_STRING)
		)
		{
			const gchar *path;
			GList *glist, *obj_list = ObjDDEBufferParsePath(
				(const guint8 *)selection_data->data,
				selection_data->length
			);
			obj_struct *obj;

			for(glist = obj_list;
			    glist != NULL;
			    glist = g_list_next(glist)
			)
			{
				obj = OBJ(glist->data);
				if(obj == NULL)
					continue;

				path = obj->path;
				if(STRISEMPTY(path))
					continue;

				paths_list = g_list_append(
					paths_list,
					STRDUP(path)
				);

				ObjDelete(obj);
			}
			g_list_free(obj_list);
		}
		/* Scan Item */
		else if(info == AVS_DND_INFO_SCAN_ITEM)
		{
#ifdef HAVE_LIBENDEAVOUR2
			const gulong start_time = edv_time();
#else
			const gulong start_time = (gulong)time(NULL);
#endif
			gint i;
			gchar **strv;
			const gchar *location;
			GList	*glist,
				*obj_list = ObjDDEBufferParse(
				(const guint8 *)selection_data->data,
				selection_data->length
			);
			obj_struct *obj;

			for(glist = obj_list;
				glist != NULL;
				glist = g_list_next(glist)
			)
			{
				obj = OBJ(glist->data);
				if(obj == NULL)
					continue;

				location = obj->path;
				if(STRISEMPTY(location))
					continue;

				/* Update the last scanned time of all the
				 * scan items who's location matches this
				 * location
				 */
				avs_win_scan_list_touch_by_location(
					win,
					location,
					start_time
				);

				/* Need to explode the location so we get
				 * each path
				 */
				strv = g_strsplit(location, ",", -1);
				if(strv == NULL)
					continue;

				/* Append each path from the location to the
				 * paths list
				 */
				for(i = 0; strv[i] != NULL; i++)
					paths_list = g_list_append(
						paths_list,
						g_strdup(strv[i])
					);

				g_strfreev(strv);

				ObjDelete(obj);
			}
			g_list_free(obj_list);
		}
		else
		{
			/* Unsupported data */
			gchar	*target_name = gdk_atom_name(
				selection_data->target
			),
					*type_name = gdk_atom_name(
				selection_data->type
			),
					*msg = g_strdup_printf(
"Data \"%s\" of type \"%s\",\n\
is not supported by this application.",
				target_name,
				type_name
			);
			g_free(target_name);
			g_free(type_name);
#ifdef HAVE_LIBENDEAVOUR2
			edv_play_sound_warning(core->edv_ctx);
#endif
			CDialogSetTransientFor(toplevel);
			CDialogGetResponse(
				"Unsupported Data",
				msg,
				NULL,
				CDIALOG_ICON_WARNING,
				CDIALOG_BTNFLAG_OK,
				CDIALOG_BTNFLAG_OK
			);
			CDialogSetTransientFor(NULL);
			g_free(msg);
		}
		if(paths_list == NULL)
		{
			avs_win_set_busy(win, FALSE);
			return;
		}

		/* Start the scan */
		avs_win_start_scan_process(
			win,
			NULL,			/* Use the virus database path
									 * specified in the configuration */
			paths_list,
			avs_win_current_scan_settings(win)
		);

		/* Switch to the results page */
		avs_win_switch_page(win, AVS_WIN_PAGE_RESULTS);

		/* Delete the paths list */
		g_list_foreach(paths_list, (GFunc)g_free, NULL);
		g_list_free(paths_list);
	}
	/* Scan List? */
	else if(w == win->scan_clist)
	{
		GList *obj_list;
		GtkCList *clist = GTK_CLIST(w);

		/* Get the list of objects based on the target data
		 * type
		 *
		 * String
		 */
		if((info == DND_INFO_TEXT_PLAIN) ||
		   (info == DND_INFO_TEXT_URI_LIST) ||
		   (info == DND_INFO_STRING)
		)
		{
			obj_list = ObjDDEBufferParsePath(
				(const guint8 *)selection_data->data,
				selection_data->length
			);
		}
		/* Scan Item */
		else if(info == AVS_DND_INFO_SCAN_ITEM)
		{
			obj_list = ObjDDEBufferParse(
				(const guint8 *)selection_data->data,
				selection_data->length
			);
		}
		else
		{
			/* Unsupported data */
			gchar	*target_name = gdk_atom_name(
				selection_data->target
			),
					*type_name = gdk_atom_name(
				selection_data->type
			),
					*msg = g_strdup_printf(
"Data \"%s\" of type \"%s\",\n\
is not supported by this application.",
				target_name,
				type_name
			);
			g_free(target_name);
			g_free(type_name);
#ifdef HAVE_LIBENDEAVOUR2
			edv_play_sound_warning(core->edv_ctx);
#endif
			CDialogSetTransientFor(toplevel);
			CDialogGetResponse(
				"Unsupported Data",
				msg,
				NULL,
				CDIALOG_ICON_WARNING,
				CDIALOG_BTNFLAG_OK,
				CDIALOG_BTNFLAG_OK
			);
			CDialogSetTransientFor(NULL);
			g_free(msg);

			obj_list = NULL;
		}
		if(obj_list == NULL)
		{
			avs_win_set_busy(win, FALSE);
			return;
		}

		/* Begin adding the objects to the scan list */
		gtk_clist_freeze(clist);

		/* Insert or append? */
		if(row > -1)
		{
			gint ins_row = row;
			GList *glist;
			obj_struct *obj;

			for(glist = obj_list;
				glist != NULL;
				glist = g_list_next(glist)
			)
			{
				obj = OBJ(glist->data);
				ins_row = avs_win_scan_list_insert(
					win, ins_row, ObjCopy(obj)
				);
				ObjDelete(obj);
				if(ins_row < 0)
					break;
				ins_row++;
			}
		}
		else
		{
			gint new_row;
			GList *glist;
			obj_struct *obj;

			for(glist = obj_list;
				glist != NULL;
				glist = g_list_next(glist)
			)
			{   
				obj = OBJ(glist->data);
				new_row = avs_win_scan_list_append(win, ObjCopy(obj));
				ObjDelete(obj);
				if(new_row < 0)
					break;
			}
		}
		g_list_free(obj_list);

		gtk_clist_thaw(clist);
	}
	/* Location Entry */
	else if(w == win->scan_location_entry)
	{
		gchar *location;
		GtkEntry *entry = GTK_ENTRY(w);

		/* Format the location string based on the target data
		 * type
		 *
		 * String
		 */
		if((info == DND_INFO_TEXT_PLAIN) ||
		   (info == DND_INFO_TEXT_URI_LIST) ||
		   (info == DND_INFO_STRING)
		)
		{
			gchar *s;
			GList *glist, *obj_list = ObjDDEBufferParsePath(
				(const guint8 *)selection_data->data,
				selection_data->length
			);
			obj_struct *obj;

			location = STRDUP("");
			for(glist = obj_list;
				glist != NULL;
				glist = g_list_next(glist)
			)
			{
				obj = OBJ(glist->data);
				if(obj == NULL)
					continue;

				s = g_strconcat(
					location,
					obj->path,
					(g_list_next(glist) != NULL) ? "," : "",
					NULL
				);
				g_free(location);
				location = s;

				ObjDelete(obj);
			}
			g_list_free(obj_list);
		}
		/* Scan Item */
		else if(info == AVS_DND_INFO_SCAN_ITEM)
		{
			gchar *s;
			GList *glist, *obj_list = ObjDDEBufferParse(
				(const guint8 *)selection_data->data,
				selection_data->length
			);
			obj_struct *obj;

			location = STRDUP("");
			for(glist = obj_list;
				glist != NULL;
				glist = g_list_next(glist)
			)
			{
				obj = OBJ(glist->data);
				if(obj == NULL)
					continue;

				s = g_strconcat(
					location,
					obj->path,
					(g_list_next(glist) != NULL) ? "," : "",
					NULL
				);
				g_free(location);
				location = s;

				ObjDelete(obj);
			}
			g_list_free(obj_list);
		}
		else
		{
			/* Unsupported data */
			gchar	*target_name = gdk_atom_name(
				selection_data->target
			),
					*type_name = gdk_atom_name(
				selection_data->type
			),
					*msg = g_strdup_printf(
"Data \"%s\" of type \"%s\",\n\
is not supported by this application.",
				target_name,
				type_name
			);
			g_free(target_name);
			g_free(type_name);
#ifdef HAVE_LIBENDEAVOUR2
			edv_play_sound_warning(core->edv_ctx);
#endif
			CDialogSetTransientFor(toplevel);
			CDialogGetResponse(
				"Unsupported Data",
				msg,
				NULL,
				CDIALOG_ICON_WARNING,
				CDIALOG_BTNFLAG_OK,
				CDIALOG_BTNFLAG_OK
			);
			CDialogSetTransientFor(NULL);
			g_free(msg);

			location = NULL;
		}
		if(location == NULL)
		{
			avs_win_set_busy(win, FALSE);
			return;
		}

		/* Set the location entry */
		gtk_entry_set_text(entry, location);
		gtk_entry_set_position(entry, -1);

		g_free(location);
	}
	/* Results List */
	else if(w == win->results_clist)
	{

	}

	avs_win_update_display(win);

	avs_win_set_busy(win, FALSE);
}


/*
 *	Animation Icon "drag_data_received" signal callback.
 */
void avs_win_dnd_anim_icon_drag_data_received_cb(
	GtkWidget *widget, GdkDragContext *dc,
	gint x, gint y,
	GtkSelectionData *selection_data, guint info, guint t,
	gpointer data
)
{
	AVSWin *win = AVS_WIN(data);
	if((widget == NULL) || (dc == NULL) || (win == NULL))
		return;
	
	avs_win_dnd_drag_data_received_nexus(
		win, widget, dc, info, selection_data,
		0, 0
	);                                         
}


/*
 *	Win GtkCList "drag_data_get" signal callback.
 */
void avs_win_dnd_list_drag_data_get_cb(
	GtkWidget *widget, GdkDragContext *dc,
	GtkSelectionData *selection_data, guint info, guint t,
	gpointer data
)
{
	gboolean data_sent = FALSE;
	gint buf_len = 0;
	guint8 *buf = NULL;
	GtkCList *clist;
	AVSWin *win = AVS_WIN(data);
	if((widget == NULL) || (dc == NULL) || (win == NULL))
		return;

	clist = GTK_IS_CLIST(widget) ? GTK_CLIST(widget) : NULL;
	if(clist == NULL)
		return;


	/* Handle by widget */

	/* Scan List */
	if(widget == win->scan_clist)
	{
		GList *glist;
		obj_struct *obj;

		for(glist = clist->selection;
			glist != NULL;
			glist = g_list_next(glist)
		)
		{
			obj = OBJ(gtk_clist_get_row_data(clist, (gint)glist->data));
			if(obj == NULL)
				continue;

		    buf = ObjDDEBufferAppend(
				buf, &buf_len, obj
			);
		}
	}
	/* Results List */
	else if(widget == win->results_clist)
	{
		GList *glist;
		obj_struct *obj;

		for(glist = clist->selection;
			glist != NULL;
			glist = g_list_next(glist)
		)
		{
			obj = OBJ(gtk_clist_get_row_data(clist, (gint)glist->data));
			if(obj == NULL)
				continue;

			if(obj->path != NULL)
				buf = ObjDDEBufferAppendPath(
					buf, &buf_len, obj
				);
		}
	}

	/* Got buffer to send out? */
	if(buf != NULL)
	{
		gtk_selection_data_set( 
			selection_data,
			GDK_SELECTION_TYPE_STRING,
			8,                      /* Bits Per Character */
			buf, buf_len
		);
		data_sent = TRUE;
		g_free(buf);
	}

	/* If no data was sent out then send an error response */
	if(!data_sent)
	{
		const gchar *s = "Error";
		gtk_selection_data_set(
			selection_data,
			GDK_SELECTION_TYPE_STRING,
			8,			/* Bits Per Character */
			s,			/* Data */
			STRLEN(s)		/* Length */
		); 
		data_sent = TRUE;
	}
}

/*
 *	Win GtkCList "drag_data_received" signal callback.
 */
void avs_win_dnd_list_drag_data_received_cb(
	GtkWidget *widget, GdkDragContext *dc,
	gint x, gint y,
	GtkSelectionData *selection_data, guint info, guint t,
	gpointer data
)
{
	gint row, column;
	GtkCList *clist;
	AVSWin *win = AVS_WIN(data);
	if((widget == NULL) || (dc == NULL) || (win == NULL))
		return;

	clist = GTK_IS_CLIST(widget) ? GTK_CLIST(widget) : NULL;
	if(clist == NULL)
		return;

	/* Find the row and column from the specified coordinates */
	if(!gtk_clist_get_selection_info(
		clist,
		x,    
		y - ((clist->flags & GTK_CLIST_SHOW_TITLES) ?
			clist->column_title_area.height +
			clist->column_title_area.y : 0),
		&row, &column
	))               
	{
		row = -1;
		column = 0;
	}

	avs_win_dnd_drag_data_received_nexus(
		win, widget, dc, info, selection_data,
		row, column
	);                                         
}

/*
 *	Win GtkCList "drag_data_delete" signal callback.
 */
void avs_win_dnd_list_drag_data_delete_cb(
	GtkWidget *widget, GdkDragContext *dc, gpointer data
)
{
	gint row;
	GList *glist, *obj_list;
	GtkCList *clist;
	AVSWin *win = AVS_WIN(data);
	if((widget == NULL) || (dc == NULL) || (win == NULL))
		return;

	clist = GTK_IS_CLIST(widget) ? GTK_CLIST(widget) : NULL;
	if(clist == NULL)
		return;

	if(clist->selection == NULL)
		return;

	gtk_clist_freeze(clist);

	/* Copy each selected row's data to obj_list */
	obj_list = NULL;
	for(glist = clist->selection;
		glist != NULL;
		glist = g_list_next(glist)
	)
	{
		row = (gint)glist->data;
		obj_list = g_list_append(
			obj_list,
			gtk_clist_get_row_data(clist, row)
		);
	}

	/* Delete each row who's has the data specified in the obj_list */
	for(glist = obj_list;
		glist != NULL;
		glist = g_list_next(glist)
	)
	{
		row = gtk_clist_find_row_from_data(clist, glist->data);
		if(row > -1)
			gtk_clist_remove(clist, row);
	}

	g_list_free(obj_list);

	gtk_clist_thaw(clist);
}


/*       
 *	Win GtkEntry "drag_data_received" signal callback.
 */
void avs_win_dnd_entry_drag_data_received_cb(
	GtkWidget *widget, GdkDragContext *dc,
	gint x, gint y,
	GtkSelectionData *selection_data, guint info, guint t,
	gpointer data                                         
)
{
	GtkEntry *entry;
	AVSWin *win = AVS_WIN(data);
	if((widget == NULL) || (dc == NULL) || (win == NULL))
		return;

	entry = GTK_IS_ENTRY(widget) ? GTK_ENTRY(widget) : NULL;
	if(entry == NULL)
		return;

	avs_win_dnd_drag_data_received_nexus(
		win, widget, dc, info, selection_data,
		0, 0
	);
}
